# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import numpy as np
from astropy import units as u

from itur.models.itu1144 import bilinear_2D_interpolator
from itur.utils import (dataset_dir, prepare_input_array, prepare_output_array,
                        prepare_quantity, load_data_interpolator)


def __fcn_columnar_content_reduced_liquid__(Lred, lat, lon, p):
        available_p = np.array(
            [0.1, 0.2, 0.3, 0.5, 1.0, 2.0, 3.0, 5.0, 10.0, 20.0, 30.0, 50.0,
             60.0, 70.0, 80.0, 90.0, 95.0])

        if p in available_p:
            p_below = p_above = p
            pExact = True
        else:
            pExact = False
            idx = available_p.searchsorted(p, side='right') - 1
            idx = np.clip(idx, 0, len(available_p))

            p_below = available_p[idx]
            p_above = available_p[idx + 1]

        # Compute the values of Lred_a
        Lred_a = Lred(lat, lon, p_above)
        if not pExact:
            Lred_b = Lred(lat, lon, p_below)
            Lred = Lred_b + (Lred_a - Lred_b) * (np.log(p) - np.log(p_below)) \
                / (np.log(p_above) - np.log(p_below))
            return Lred
        else:
            return Lred_a


class __ITU840__():
    """Attenuation due to clouds and fog: This Recommendation provides methods
    to predict the attenuation due to clouds and fog on Earth-space paths.

    Available versions include:
    * P.840-4 (10/09) (Superseded)
    * P.840-5 (02/12) (Superseded)
    * P.840-6 (09/13) (Superseded)
    * P.840-7 (12/17) (Superseded)
    * P.840-8 (08/19) (Current version)

    Non-available versions include:
    * P.840-1 (08/94) (Superseded) - Tentative similar to P.840-4
    * P.840-2 (08/97) (Superseded) - Tentative similar to P.840-4
    * P.840-3 (10/99) (Superseded) - Tentative similar to P.840-4

    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.840 recommendation.

    def __init__(self, version=7):
        if version == 8:
            self.instance = _ITU840_8_()
        elif version == 7:
            self.instance = _ITU840_7_()
        elif version == 6:
            self.instance = _ITU840_6_()
        elif version == 5:
            self.instance = _ITU840_5_()
        elif version == 4:
            self.instance = _ITU840_4_()
        else:
            raise ValueError(
                'Version {0}  is not implemented for the ITU-R P.840 model.'
                .format(version))

    @property
    def __version__(self):
        return self.instance.__version__

    def specific_attenuation_coefficients(self, f, T):
        # Abstract method to compute the specific attenuation coefficients
        fcn = np.vectorize(self.instance.specific_attenuation_coefficients)
        return fcn(f, T)

    def columnar_content_reduced_liquid(self, lat, lon, p):
        # Abstract method to compute the columnar content of reduced liquid
        fcn = np.vectorize(__fcn_columnar_content_reduced_liquid__,
                           excluded=[0, 1, 2], otypes=[np.ndarray])
        return np.array(fcn(self.instance.Lred, lat, lon, p).tolist())

    def cloud_attenuation(self, lat, lon, el, f, p, Lred=None):
        # Abstract method to compute the cloud attenuation
        Kl = self.specific_attenuation_coefficients(f, T=0)
        if Lred is None:
            Lred = self.columnar_content_reduced_liquid(lat, lon, p)
        A = Lred * Kl / np.sin(np.deg2rad(el))

        return A

    def lognormal_approximation_coefficient(self, lat, lon):
        # Abstract method to compute the lognormal approximation coefficients
        return self.instance.lognormal_approximation_coefficient(lat, lon)


class _ITU840_8_():

    def __init__(self):
        self.__version__ = 8
        self.year = 2019
        self.month = 8
        self.link = 'https://www.itu.int/rec/R-REC-P.840-8-201908-I/en'

        self._Lred = {}
        self._M = None
        self._sigma = None
        self._Pclw = None

    # Note: The dataset used in recommendation 840-8 is the same as the
    # dataset use in recommendation 840-7. (The zip files included in
    # both recommendations are identical)

    def Lred(self, lat, lon, p):
        if not self._Lred:
            ps = [0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10, 20, 30,
                  50, 60, 70, 80, 90, 95]
            d_dir = os.path.join(dataset_dir, '840/v7_lred_%s.npz')
            for p_load in ps:
                self._Lred[float(p_load)] = load_data_interpolator(
                       '840/v7_lat.npz', '840/v7_lon.npz',
                       d_dir % (str(p_load).replace('.', '')),
                       bilinear_2D_interpolator, flip_ud=False)

        return self._Lred[float(p)](
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def M(self, lat, lon):
        if not self._M:
            self._M = load_data_interpolator(
                '840/v7_lat.npz', '840/v7_lon.npz',
                '840/v7_m.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._M(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def sigma(self, lat, lon):
        if not self._sigma:
            self._sigma = load_data_interpolator(
                '840/v7_lat.npz', '840/v7_lon.npz',
                '840/v7_sigma.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._sigma(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def Pclw(self, lat, lon):
        if not self._Pclw:
            self._Pclw = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v6_pclw.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._Pclw(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    @staticmethod
    def specific_attenuation_coefficients(f, T):
        """
        """
        return _ITU840_6_.specific_attenuation_coefficients(f, T)

    def lognormal_approximation_coefficient(self, lat, lon):
        m = self.M(lat, lon)
        sigma = self.sigma(lat, lon)
        Pclw = self.Pclw(lat, lon)

        return m, sigma, Pclw


class _ITU840_7_():

    def __init__(self):
        self.__version__ = 7
        self.year = 2017
        self.month = 12
        self.link = 'https://www.itu.int/rec/R-REC-P.840-7-201712-I/en'

        self._Lred = {}
        self._M = {}
        self._sigma = {}
        self._Pclw = {}

    def Lred(self, lat, lon, p):
        if not self._Lred:
            ps = [0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10, 20, 30,
                  50, 60, 70, 80, 90, 95]
            d_dir = os.path.join(dataset_dir, '840/v7_lred_%s.npz')
            for p_load in ps:
                self._Lred[float(p_load)] = load_data_interpolator(
                       '840/v7_lat.npz', '840/v7_lon.npz',
                       d_dir % (str(p_load).replace('.', '')),
                       bilinear_2D_interpolator, flip_ud=False)

        return self._Lred[float(p)](
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def M(self, lat, lon):
        if not self._M:
            self._M = load_data_interpolator(
                '840/v7_lat.npz', '840/v7_lon.npz',
                '840/v7_m.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._M(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def sigma(self, lat, lon):
        if not self._sigma:
            self._sigma = load_data_interpolator(
                '840/v7_lat.npz', '840/v7_lon.npz',
                '840/v7_sigma.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._sigma(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def Pclw(self, lat, lon):
        if not self._Pclw:
            self._Pclw = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v6_pclw.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._Pclw(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    @staticmethod
    def specific_attenuation_coefficients(f, T):
        """
        """
        return _ITU840_6_.specific_attenuation_coefficients(f, T)

    def lognormal_approximation_coefficient(self, lat, lon):
        #TODO: This is the wrong method, Need to update
        m = self.M(lat, lon)
        sigma = self.sigma(lat, lon)
        Pclw = self.Pclw(lat, lon)

        return m, sigma, Pclw


class _ITU840_6_():

    def __init__(self):
        self.__version__ = 6
        self.year = 2013
        self.month = 9
        self.link = 'https://www.itu.int/rec/R-REC-P.840-6-201202-I/en'

        self._Lred = {}
        self._M = {}
        self._sigma = {}
        self._Pclw = {}

    def Lred(self, lat, lon, p):
        if not self._Lred:
            ps = [0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10, 20, 30,
                  50, 60, 70, 80, 90, 95]
            d_dir = os.path.join(dataset_dir, '840/v6_lred_%s.npz')
            for p_load in ps:
                self._Lred[float(p_load)] = load_data_interpolator(
                       '840/v6_lat.npz', '840/v6_lon.npz',
                       d_dir % (str(p_load).replace('.', '')),
                       bilinear_2D_interpolator, flip_ud=False)

        return self._Lred[float(p)](
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def M(self, lat, lon):
        if not self._M:
            self._M = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v6_m.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._M(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def sigma(self, lat, lon):
        if not self._sigma:
            self._sigma = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v6_sigma.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._sigma(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def Pclw(self, lat, lon):
        if not self._Pclw:
            self._Pclw = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v6_pclw.npz', bilinear_2D_interpolator, flip_ud=False)

        return self._Pclw(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    @staticmethod
    def specific_attenuation_coefficients(f, T):
        """
        """
        if np.any(f > 1000):
            raise ValueError('Frequency must be introduced in GHz and the '
                             'maximum range is 1000 GHz')

        T_kelvin = T + 273.15
        theta = 300.0 / T_kelvin                # Eq. 9

        # Compute the values of the epsilons
        epsilon0 = 77.66 + 103.3 * (theta - 1)  # Eq. 6
        epsilon1 = 0.0671 * epsilon0            # Eq. 7
        epsilon2 = 3.52                         # Eq. 8

        # Compute the principal and secondary relacation frequencies
        fp = 20.20 - 146 * (theta - 1) + 316.0 * (theta - 1)**2     # Eq. 10
        fs = 39.8 * fp                                              # Eq. 11

        # Compute the dielectric permitivity of water
        epsilonp = (epsilon0 - epsilon1) / (1 + (f / fp) ** 2) + \
            (epsilon1 - epsilon2) / (1 + (f / fs) ** 2) + epsilon2  # Eq. 5

        epsilonpp = f * (epsilon0 - epsilon1) / (fp * (1 + (f / fp)**2)) + \
            f * (epsilon1 - epsilon2) / (fs * (1 + (f / fs)**2))       # Eq. 4

        eta = (2 + epsilonp) / epsilonpp                    # Eq. 3
        Kl = (0.819 * f) / (epsilonpp * (1 + eta**2))       # Eq. 2

        return Kl       # Specific attenuation coefficient  (dB/km)/(g/m3)

    def lognormal_approximation_coefficient(self, lat, lon):
        m = self.M(lat, lon)
        sigma = self.sigma(lat, lon)
        Pclw = self.Pclw(lat, lon)

        return m, sigma, Pclw


class _ITU840_5_():

    def __init__(self):
        self.__version__ = 5
        self.year = 2012
        self.month = 2
        self.link = 'https://www.itu.int/rec/R-REC-P.840-5-201202-S/en'

        self._Lred = {}
        self._M = {}
        self._sigma = {}
        self._Pclw = {}

    def Lred(self, lat, lon, p):
        if not self._Lred:
            ps = [0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10, 20, 30,
                  50, 60, 70, 80, 90, 95]
            d_dir = os.path.join(dataset_dir, '840/v4_esawred_%s.npz')
            for p_load in ps:
                self._Lred[float(p_load)] = load_data_interpolator(
                       '840/v4_lat.npz', '840/v4_lon.npz',
                       d_dir % (str(p_load).replace('.', '')),
                       bilinear_2D_interpolator, flip_ud=False)

        return self._Lred[float(p)](
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def M(self, lat, lon):
        if not self._M:
            self._M = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v4_wred_lognormal_mean.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._M(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def sigma(self, lat, lon):
        if not self._sigma:
            self._sigma = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v4_wred_lognormal_stdev.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._sigma(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def Pclw(self, lat, lon):
        if not self._Pclw:
            self._Pclw = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v4_wred_lognormal_pclw.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._Pclw(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    @staticmethod
    def specific_attenuation_coefficients(f, T):
        """
        """
        return _ITU840_4_.specific_attenuation_coefficients(f, T)

    def lognormal_approximation_coefficient(self, lat, lon):
        m = self.M(lat, lon)
        sigma = self.sigma(lat, lon)
        Pclw = self.Pclw(lat, lon)

        return m, sigma, Pclw


class _ITU840_4_():

    def __init__(self):
        self.__version__ = 4
        self.year = 2013
        self.month = 9
        self.link = 'https://www.itu.int/rec/R-REC-P.840-6-201202-I/en'

        self._Lred = {}
        self._M = {}
        self._sigma = {}
        self._Pclw = {}

    def Lred(self, lat, lon, p):
        if not self._Lred:
            ps = [0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10, 20, 30,
                  50, 60, 70, 80, 90, 95]
            d_dir = os.path.join(dataset_dir, '840/v4_esawred_%s.npz')
            for p_load in ps:
                self._Lred[float(p_load)] = load_data_interpolator(
                       '840/v4_lat.npz', '840/v4_lon.npz',
                       d_dir % (str(p_load).replace('.', '')),
                       bilinear_2D_interpolator, flip_ud=False)

        return self._Lred[float(p)](
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def M(self, lat, lon):
        if not self._M:
            self._M = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v4_wred_lognormal_mean.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._M(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def sigma(self, lat, lon):
        if not self._sigma:
            self._sigma = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v4_wred_lognormal_stdev.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._sigma(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    def Pclw(self, lat, lon):
        if not self._Pclw:
            self._Pclw = load_data_interpolator(
                '840/v6_lat.npz', '840/v6_lon.npz',
                '840/v4_wred_lognormal_pclw.npz', bilinear_2D_interpolator,
                flip_ud=False)

        return self._Pclw(
            np.array([lat.ravel(), lon.ravel()]).T).reshape(lat.shape)

    @staticmethod
    def specific_attenuation_coefficients(f, T):
        """
        """
        if np.any(f > 1000):
            raise ValueError(
                'Frequency must be introduced in GHz and the maximum range'
                ' is 1000 GHz')

        T_kelvin = T + 273.15
        theta = 300.0 / T_kelvin                # Eq. 9

        # Compute the values of the epsilons
        epsilon0 = 77.66 + 103.3 * (theta - 1)  # Eq. 6
        epsilon1 = 5.48                         # Eq. 7
        epsilon2 = 3.51                         # Eq. 8

        # Compute the principal and secondary relacation frequencies
        fp = 20.09 - 142 * (theta - 1) + 294.0 * (theta - 1)**2     # Eq. 10
        fs = 590 - 1500 * (theta - 1)                               # Eq. 11

        # Compute the dielectric permitivity of water
        epsilonp = (epsilon0 - epsilon1) / (1 + (f / fp) ** 2) + \
            (epsilon1 - epsilon2) / (1 + (f / fs) ** 2) + epsilon2  # Eq. 5

        epsilonpp = f * (epsilon0 - epsilon1) / (fp * (1 + (f / fp)**2)) + \
            f * (epsilon1 - epsilon2) / (fs * (1 + (f / fs)**2))       # Eq. 4

        eta = (2 + epsilonp) / epsilonpp                    # Eq. 3
        Kl = (0.819 * f) / (epsilonpp * (1 + eta**2))       # Eq. 2

        return Kl       # Specific attenuation coefficient  (dB/km)/(g/m3)

    def lognormal_approximation_coefficient(self, lat, lon):
        m = self.M(lat, lon)
        sigma = self.sigma(lat, lon)
        Pclw = self.Pclw(lat, lon)

        return m, sigma, Pclw


__model = __ITU840__()


def change_version(new_version):
    """
    Change the version of the ITU-R P.840 recommendation currently being used.


    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:
          * 8: Activates recommendation ITU-R P.840-8 (08/19) (Current version)
          * 7: Activates recommendation ITU-R P.840-7 (12/17) (Superseded)
          * 6: Activates recommendation ITU-R P.840-6 (09/13) (Superseded)
          * 5: Activates recommendation ITU-R P.840-5 (02/12) (Superseded)
          * 4: Activates recommendation ITU-R P.840-4 (10/09) (Superseded)
    """
    global __model
    __model = __ITU840__(new_version)


def get_version():
    """
    Obtain the version of the ITU-R P.840 recommendation currently being used.

    Returns
    -------
    version: int
        Version currently being used.
    """
    return __model.__version__


def specific_attenuation_coefficients(f, T):
    """
    Compute the specific attenuation coefficient for cloud attenuation.

    A method to compute the specific attenuation coefficient. The method is
    based on Rayleigh scattering, which uses a double-Debye model for the
    dielectric permittivity of water.

    This model can be used to calculate the value of the specific attenuation
    coefficient for frequencies up to 1000 GHz:


    Parameters
    ----------
    f : number
        Frequency (GHz)
    T : number
        Temperature (degrees C)


    Returns
    -------
    Kl: numpy.ndarray
        Specific attenuation coefficient (dB/km)


    References
    ----------
    [1] Attenuation due to clouds and fog:
    https://www.itu.int/rec/R-REC-P.840/en
    """
    f = prepare_quantity(f, u.GHz, 'Frequency')
    T = prepare_quantity(T, u.deg_C, 'Temperature')
    return __model.specific_attenuation_coefficients(f, T)


def columnar_content_reduced_liquid(lat, lon, p):
    """
    Compute the total columnar contents of reduced cloud liquid water.

    A method to compute the total columnar content of reduced cloud liquid
    water, Lred (kg/m2), exceeded for p% of the average year


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    p : number
        Percentage of time exceeded for p% of the average year


    Returns
    -------
    Lred: numpy.ndarray
        Total columnar content of reduced cloud liquid water, Lred (kg/m2),
        exceeded for p% of the average year



    References
    ----------
    [1] Attenuation due to clouds and fog:
    https://www.itu.int/rec/R-REC-P.840/en
    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.columnar_content_reduced_liquid(lat, lon, p)
    return prepare_output_array(val, type_output) * u.kg / u.m**2


def cloud_attenuation(lat, lon, el, f, p, Lred=None):
    """
    Compute the cloud attenuation in a slant path.

    A method to estimate the attenuation due to clouds along slant paths for
    a given probability. If local measured data of the total columnar content
    of cloud liquid water reduced to a temperature of 273.15 K, Lred, is
    available from other sources, (e.g., from ground radiometric measurements,
    Earth observation products, or meteorological numerical products), the
    value should be used directly.

    The value of the cloud attenuation is computed as:

    .. math::
      A=\\frac{L_{red}(\\text{lat}, \\text{lon}, p, T) \\cdot K_l(f, T)}{\\sin(\\text{el})}


    where:
        * :math:`L_{red}` : total columnar content of liquid water reduced to a
          temperature of 273.15 K (kg/m2);
        * :math:`K_l` : specific attenuation coefficient ((dB/km)/(g/m3));
        * :math:`el` : path elevation angle (deg).
        * :math:`f` : frequency (GHz).
        * :math:`p` : Percentage of time exceeded for p% of the average year (%).
        * :math:`T` : temperature (K). Equal to 273.15 K.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    el : number, sequence, or numpy.ndarray
        Elevation angle of the receiver points (deg)
    f : number
        Frequency (GHz)
    p : number
         Percentage of time exceeded for p% of the average year
    Lred: number
        Total columnar contents of reduced cloud liquid water. (kg/m2)



    Returns
    -------
    A: numpy.ndarray
        Cloud attenuation, A (dB), exceeded for p% of the average year



    References
    ----------
    [1] Attenuation due to clouds and fog:
    https://www.itu.int/rec/R-REC-P.840/en
    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    el = prepare_quantity(el, u.deg, 'Elevation angle')
    f = prepare_quantity(f, u.GHz, 'Frequency')
    Lred = prepare_quantity(
      Lred, u.kg / u.m**2,
      'Total columnar contents of reduced cloud liquid water.')
    val = __model.cloud_attenuation(lat, lon, el, f, p, Lred)
    return prepare_output_array(val, type_output) * u.dB


def lognormal_approximation_coefficient(lat, lon):
    """
    Total columnar contents of cloud liquid water distribution coefficients.

    The annual statistics of the total columnar content of reduced cloud
    liquid water content can be approximated by a log-normal distribution.
    This function computes the coefficients for the mean, :math:`m`,
    standard deviation, :math:`\sigma`, and probability of non-zero reduced
    total columnar content of cloud liquid water, :math:`Pclw`, for such the
    log-normal distribution.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points


    Returns
    -------
    m: numpy.ndarray
        Mean of the lognormal distribution
    σ: numpy.ndarray
        Standard deviation of the lognormal distribution
    Pclw: numpy.ndarray
        Probability of cloud liquid water of the lognormal distribution



    References
    ----------
    [1] Attenuation due to clouds and fog:
    https://www.itu.int/rec/R-REC-P.840/en
    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    val = __model.lognormal_approximation_coefficient(lat, lon)
    u_adim = u.dimensionless_unscaled
    return (prepare_output_array(val[0], type_output) * u_adim,
            prepare_output_array(val[1], type_output) * u_adim,
            prepare_output_array(val[2], type_output) * u_adim)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import scipy.stats as stats
from scipy.signal import lfilter

from itur.models.itu618 import rain_attenuation, scintillation_attenuation_sigma
from itur.models.itu676 import gamma0_exact, \
    slant_inclined_path_equivalent_height
from itur.models.itu840 import lognormal_approximation_coefficient, \
    specific_attenuation_coefficients
from itur.models.itu835 import standard_pressure, standard_water_vapour_density
from itur.models.itu836 import total_water_vapour_content
from itur.models.itu837 import rainfall_probability
from itur.models.itu676 import zenit_water_vapour_attenuation
from itur.models.itu1510 import surface_mean_temperature
from itur.models.itu1511 import topographic_altitude
from itur.utils import prepare_quantity

from astropy import units as u


class __ITU1853():

    """Tropospheric attenuation time series synthesis

    Available versions include:
    * P.1853-0 (10/09) (Superseded)
    * P.1853-1 (02/12) (Current version)
    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.1853 recommendation.

    def __init__(self, version=1):
        if version == 1:
            self.instance = _ITU1853_1()
        elif version == 0:
            self.instance = _ITU1853_0()
        else:
            raise ValueError(
                'Version {0} is not implemented for the ITU-R P.1853 model.'
                .format(version))

    @staticmethod
    def set_seed(seed):
        np.random.seed(seed)

    @property
    def __version__(self):
        return self.instance.__version__

    def rain_attenuation_synthesis(self, lat, lon, f, el, hs, Ns,
                                   Ts=1, tau=45, n=None):
        return self.instance.rain_attenuation_synthesis(
                lat, lon, f, el, hs, Ns, Ts=Ts, tau=tau, n=n)

    def total_attenuation_synthesis(self, lat, lon, f, el, p, D, Ns, Ts=1,
                                    tau=45, hs=None, eta=0.65, rho=None,
                                    H=None, P=None, hL=1000,
                                    return_contributions=False):
        return self.instance.total_attenuation_synthesis(
            lat, lon, f, el, p, D, Ns, Ts=Ts, tau=tau, hs=hs, eta=eta, rho=rho,
            H=H, P=P, hL=hL, return_contributions=return_contributions)

    def scintillation_attenuation_synthesis(self, Ns, f_c=0.1, Ts=1):
        return self.instance.scintillation_attenuation_synthesis(
                Ns, f_c=f_c, Ts=Ts)

    def cloud_liquid_water_synthesis(self, lat, lon, Ns, Ts=1, n=None):
        return self.instance.cloud_liquid_water_synthesis(
                lat, lon, Ns, Ts=Ts, n=n)

    def integrated_water_vapour_synthesis(self, lat, lon, Ns, Ts=1, n=None):
        return self.instance.integrated_water_vapour_synthesis(
                lat, lon, Ns, Ts=Ts, n=n)


class _ITU1853_1():

    def __init__(self):
        self.__version__ = 1
        self.year = 2012
        self.month = 2
        self.link = 'https://www.itu.int/rec/R-REC-P.1853-1-201202-I/en'

    @staticmethod
    def rain_attenuation_synthesis(
            lat, lon, f, el, hs, Ns, tau=45, Ts=1, n=None):
        """
        For Earth-space paths, the time series synthesis method is valid for
        frequencies between 4 GHz and 55 GHz and elevation angles between
        5 deg and 90 deg.
        """

        # Step A1: Determine Prain (% of time), the probability of rain on the
        # path. Prain can be well approximated as P0(lat, lon)
        P_rain = rainfall_probability(lat, lon).\
            to(u.dimensionless_unscaled).value

        # Step A2: Construct the set of pairs [Pi, Ai] where Pi (% of time) is
        # the probability the attenuation Ai (dB) is exceeded where Pi < P_K
        p_i = np.array([0.01, 0.02, 0.03, 0.05,
                        0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10])
        Pi = np.array([p for p in p_i if p < P_rain * 100], dtype=np.float)
        Ai = np.array([0 for p in p_i if p < P_rain * 100], dtype=np.float)

        for i, p in enumerate(Pi):
            Ai[i] = rain_attenuation(lat, lon, f, el, hs, p, tau=tau).value

        # Step A3: Transform the set of pairs [Pi, Ai] to [Q^{-1}(Pi/P_k),
        # ln(Ai)]
        Q = stats.norm.ppf((Pi / 100))
        lnA = np.log(Ai)

        # Step A4: Determine the variables sigma_lna, m_lna by performing a
        # least-squares fit to lnAi = sigma_lna Q^{-1}(Pi/P_k) + m_lna
        m, sigma = np.linalg.lstsq(np.vstack([np.ones(len(Q)), Q]).T,
                                   lnA, rcond=None)[0]

        # Step B: Set the low-pass filter parameter
        beta = 2e-4
        # Step C: compute the attenuation offset
        A_offset = np.exp(m + sigma * stats.norm.ppf(P_rain))
        # Step D: Time series synthesis
        # D1: Synthesize a white Gaussian noise time series
        if n is None:
            n = np.random.normal(0, 1, int(Ns * Ts + 2e5))[::Ts]
            discard_samples = True
        else:
            discard_samples = False

        # D2, D3 : Filter the noise time series with a recursive low-pass
        # filter
        rho = np.exp(-beta * Ts)
        X = lfilter([np.sqrt(1 - rho**2)], [1, -rho], n, 0)
        # D4: Compute Y_rain
        Y_rain = np.exp(m + sigma * X)
        # D5: Compute Arain
        A_rain = np.maximum(Y_rain - A_offset, 0)

        # D6: Discard the first 200 000 samples from the synthesized
        if discard_samples:
            A_rain = A_rain[np.ceil(200000/Ts).astype(int):]

        return A_rain.flatten()

    @staticmethod
    def fftnoise(f):
        f = np.array(f, dtype='complex')
        Np = (len(f) - 1) // 2
        phases = np.random.rand(Np) * 2 * np.pi
        phases = np.cos(phases) + 1j * np.sin(phases)
        f[1:Np + 1] *= phases
        f[-1:-1 - Np:-1] = np.conj(f[1:Np + 1])
        return np.fft.ifft(f).real

    def scintillation_attenuation_synthesis(self, Ns, f_c=0.1, Ts=1):
        """
        For Earth-space paths, the time series synthesis method is valid for
        frequencies between 4 GHz and 55 GHz and elevation angles between
        5 deg and 90 deg.
        """
        freqs = np.abs(np.fft.fftfreq(2 * int(Ns + 2e5), 1 / Ts))
        H_f = np.where(freqs <= f_c, 1, 10 **
                       ((np.log10(freqs) - np.log10(f_c)) * (-8 / 3)))
        H_f = H_f[0:int(Ns + 2e5)]
        sci = self.fftnoise(np.fft.fftshift(H_f))
        return sci[200000:].flatten()

    @staticmethod
    def cloud_liquid_water_synthesis(lat, lon, Ns, Ts=1, n=None):
        """
        """

        # Step A: Estimation of m, sigma and Pcwl
        m, sigma, Pcwl = lognormal_approximation_coefficient(lat, lon)
        m = m.value
        sigma = sigma.value
        Pcwl = Pcwl.value / 100

        # Step B: Low pass filter parameters
        beta_1 = 7.17e-4
        beta_2 = 2.01e-5
        gamma_1 = 0.349
        gamma_2 = 0.830

        # Step C: Truncation threshold
        alpha = stats.norm.ppf(1 - Pcwl)

        # Step D: Time series synthesis
        # Step D1: Synthesize a white Gaussian noise time series
        if n is None:
            n = np.random.normal(0, 1, int(Ns * Ts + 5e5))[::Ts]
            discard_samples = True
        else:
            discard_samples = False

        # Step D3: Filter the noise time series, with two recursive low-pass
        # filters
        rho_1 = np.exp(-beta_1 * Ts)
        X_1 = lfilter([np.sqrt(1 - rho_1**2)], [1, -rho_1], n, 0)
        rho_2 = np.exp(-beta_2 * Ts)
        X_2 = lfilter([np.sqrt(1 - rho_2**2)], [1, -rho_2], n, 0)
        # Step D4: Compute Gc(kTs),
        G_c = gamma_1 * X_1 + gamma_2 * X_2
        # Step D5: Compute L(kTs) (dB)
        L = np.where(G_c > alpha, np.exp(m + sigma * stats.norm.ppf(
            1 - 1 / Pcwl * stats.norm.sf(G_c))), 0)

        # D6: Discard the first 500 000 samples from the synthesized
        if discard_samples:
            L = L[np.ceil(500000/Ts).astype(int):]

        return L.flatten()

    @staticmethod
    def integrated_water_vapour_coefficients(lat, lon):
        # A Estimation of κ and λ
        ps = np.array([0.1, 0.2, 0.3, 0.5, 1, 2, 3, 5, 10, 20, 30, 50])
        Vi = np.array([total_water_vapour_content(lat, lon, p_i).value
                       for p_i in ps])

        ln_lnPi = np.log(- np.log(ps / 100))
        ln_Vi = np.log(Vi)

        a, b = np.linalg.lstsq(np.vstack([ln_Vi, np.ones(len(ln_Vi))]).T,
                               ln_lnPi)[0]
        kappa = a
        lambd = np.exp(-b / a)
        return kappa, lambd

    def integrated_water_vapour_synthesis(self, lat, lon, Ns, Ts=1, n=None):
        # A Estimation of κ and λ
        kappa, lambd = self.integrated_water_vapour_coefficients(lat, lon)

        # B Low-pass filter parameter
        beta_V = 3.24e-6

        # Step C: Time series synthesis
        # Step C1: Synthesize a white Gaussian noise time series
        if n is None:
            n = np.random.normal(0, 1, (Ns * Ts + 5000000))[::Ts]
            discard_samples = True
        else:
            discard_samples = False

        # Step C3: Filter the noise time series, with two recursive low-pass
        # filters
        rho = np.exp(-beta_V * Ts)
        G_v = lfilter([np.sqrt(1 - rho**2)], [1, -rho], n, 0)
        # Step C4: Compute Compute V(kTs),
        V = lambd * (- np.log10(stats.norm.sf(G_v)))**(1 / kappa)
        # Step C5: Discard the first 5 000 000 samples from the synthesized
        if discard_samples:
            V = V[np.ceil(5000000/Ts).astype(int):]

        return V.flatten()

    def total_attenuation_synthesis(self, lat, lon, f, el, p, D, Ns, Ts=1,
                                    hs=None, tau=45, eta=0.65, rho=None,
                                    H=None, P=None, hL=1000,
                                    return_contributions=False):
        t_disc = int(5e6)
        # Step A Correlation coefficients:
        C_RC = 1
        C_CV = 0.8

        # Step B Scintillation polynomials
        def a_Fade(p):
            return -0.061 * np.log10(p)**3 + 0.072 * \
                np.log10(p)**2 - 1.71 * np.log10(p) + 3

        def a_Enhanc(p):
            return -0.0597 * np.log10(p)**3 - 0.0835 * \
                np.log10(p)**2 - 1.258 * np.log10(p) + 2.672

        # Step C1-C3:
        n_R = np.random.normal(0, 1, int((Ns * Ts + t_disc)))
        n_L0 = np.random.normal(0, 1, int((Ns * Ts + t_disc)))
        n_V0 = np.random.normal(0, 1, int((Ns * Ts + t_disc)))

        # Step C4-C5:
        n_L = C_RC * n_R + np.sqrt(1 - C_RC**2) * n_L0
        n_V = C_CV * n_L + np.sqrt(1 - C_CV**2) * n_V0

        # Step C6: Compute the rain attenuation time series
        if hs is None:
            hs = topographic_altitude(lat, lon)
        Ar = self.rain_attenuation_synthesis(
            lat, lon, f, el, hs, Ns, Ts=1, tau=tau, n=n_R)
        Ar = Ar[t_disc:]

        # Step C7: Compute the cloud integrated liquid water content time
        # series
        L = self.cloud_liquid_water_synthesis(lat, lon, Ns, Ts=1, n=n_L)
        L = L[t_disc:]
        Ac = L * \
            specific_attenuation_coefficients(f, T=0) / np.sin(np.deg2rad(el))
        Ac = Ac.flatten()

        # Step C9: Identify time stamps where A_R > 0 L > 1
        idx = np.where(np.logical_and(Ar > 0, L > 1))[0]
        idx_no = np.where(np.logical_not(
                    np.logical_and(Ar > 0, L > 1)))[0]

        # Step C10: Discard the previous values of Ac and re-compute them by
        # linear interpolation vs. time starting from the non-discarded cloud
        # attenuations values
        Ac[idx] = np.interp(idx, idx_no, Ac[idx_no])

        # Step C11: Compute the integrated water vapour content time series
        V = self.integrated_water_vapour_synthesis(lat, lon, Ns, Ts=1, n=n_V)
        V = V[t_disc:]

        # Step C12: Convert the integrated water vapour content time series
        # V into water vapour attenuation time series AV(kTs)
        Av = zenit_water_vapour_attenuation(lat, lon, p, f, V_t=V).value

        # Step C13: Compute the mean annual temperature Tm for the location of
        # interest using experimental values if available.
        Tm = surface_mean_temperature(lat, lon).value

        # Step C14: Convert the mean annual temperature Tm into mean annual
        # oxygen attenuation AO following the method recommended in
        # Recommendation ITU-R P.676.
        if P is None:
            P = standard_pressure(hs).value

        if rho is None:
            rho = standard_water_vapour_density(hs).value

        e = Tm * rho / 216.7
        go = gamma0_exact(f, P, rho, Tm).value
        ho, _ = slant_inclined_path_equivalent_height(f, P + e).value
        Ao = ho * go * np.ones_like(Ar)

        # Step C15: Synthesize unit variance scintillation time series
        sci_0 = self.scintillation_attenuation_synthesis(Ns * Ts, Ts=1)

        # Step C16: Compute the correction coefficient time series Cx(kTs) in
        # order to distinguish between scintillation fades and enhancements:
        Q_sci = 100 * stats.norm.sf(sci_0)
        C_x = np.where(sci_0 > 0, a_Fade(Q_sci) / a_Enhanc(Q_sci), 1)

        # Step C17: Transform the integrated water vapour content time series
        # V(kTs) into the Gamma distributed time series Z(kTs) as follows:
        kappa, lambd = self.integrated_water_vapour_coefficients(lat, lon)
        Z = stats.gamma.ppf(1 - np.exp(-(V / lambd)**kappa), 10, 0.1)

        # Step C18: Compute the scintillation standard deviation σ following
        # the method recommended in Recommendation ITU-R P.618.
        sigma = scintillation_attenuation_sigma(lat, lon, f, el, p, D, eta, Tm,
                                                H, P, hL).value

        # Step C19: Compute the scintillation time series sci:
        As = np.where(Ar > 1, sigma * sci_0 * C_x * Z * Ar ** (5 / 12),
                      sigma * sci_0 * C_x * Z)

        # Step C20: Compute total tropospheric attenuation time series A(kTs)
        # as follows:
        A = Ar + Ac + Av + Ao + As

        if return_contributions:
            return (Ao + Av)[::Ts], Ac[::Ts], Ar[::Ts], As[::Ts], A[::Ts]
        else:
            return A[::Ts]


class _ITU1853_0():

    def __init__(self):
        self.__version__ = 0
        self.year = 2009
        self.month = 10
        self.link = 'https://www.itu.int/rec/R-REC-P.1853-0-200910-I/en'

    @staticmethod
    def rain_attenuation_synthesis(*args, **kwargs):
        return _ITU1853_1.rain_attenuation_synthesis(*args, **kwargs)

    @staticmethod
    def scintillation_attenuation_synthesis(*args, **kwargs):
        return _ITU1853_1.scintillation_attenuation_synthesis(*args, **kwargs)

    @staticmethod
    def cloud_liquid_water_synthesis(*args, **kwargs):
        raise NotImplementedError(
            "Recommendation ITU-R P.1853 does not specify a method to compute "
            "time series for the cloud liquid water content.")

    @staticmethod
    def integrated_water_vapour_synthesis(*args, **kwargs):
        raise NotImplementedError(
            "Recommendation ITU-R P.1853 does not specify a method to compute "
            "time series for the water vapour content.")

    @staticmethod
    def total_attenuation_synthesis(*args, **kwargs):
        raise NotImplementedError(
            "Recommendation ITU-R P.1853 does not specify a method to compute "
            "time series for the total atmospheric attenuation.")


__model = __ITU1853()


def change_version(new_version):
    """
    Change the version of the ITU-R P.1853 recommendation currently being used.


    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:
          * 1:  Activates recommendation ITU-R P.1853-1 (02/12) (Current version)
    """
    global __model
    __model = __ITU1853(new_version)


def get_version():
    """
    Obtain the version of the ITU-R P.1853 recommendation currently being used.
    """
    global __model
    return __model.__version__


def rain_attenuation_synthesis(lat, lon, f, el, hs, Ns, Ts=1, tau=45, n=None):
    """
    A method to generate a synthetic time series of rain attenuation values.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    f : number or Quantity
        Frequency (GHz)
    el : sequence, or number
        Elevation angle (degrees)
    hs : number, sequence, or numpy.ndarray, optional
        Heigh above mean sea level of the earth station (km). If local data for
        the earth station height above mean sea level is not available, an
        estimate is obtained from the maps of topographic altitude
        given in Recommendation ITU-R P.1511.
    Ns : int
        Number of samples
    Ts : int
        Time step between consecutive samples (seconds)
    tau : number, optional
        Polarization tilt angle relative to the horizontal (degrees)
        (tau = 45 deg for circular polarization). Default value is 45
    n : list, np.array, optional
        Additive White Gaussian Noise used as input for the


    Returns
    -------
    rain_att: numpy.ndarray
        Synthesized rain attenuation time series (dB)


    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.1853/en
    """
    global __model

    lon = np.mod(lon, 360)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(el, u.deg, 'Elevation angle')
    hs = prepare_quantity(
        hs, u.km, 'Heigh above mean sea level of the earth station')
    Ts = prepare_quantity(f, u.second, 'Time step between samples')
    val = __model.rain_attenuation_synthesis(lat, lon, f, el, hs, Ns,
                                             Ts=Ts, tau=tau, n=n)
    return val * u.dB


def scintillation_attenuation_synthesis(Ns, f_c=0.1, Ts=1):
    """
    A method to generate a synthetic time series of scintillation attenuation
    values.


    Parameters
    ----------
    Ns : int
        Number of samples
    f_c : float
        Cut-off frequency for the low pass filter
    Ts : int
        Time step between consecutive samples (seconds)

    Returns
    -------
    sci_att: numpy.ndarray
        Synthesized scintilation attenuation time series (dB)


    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.1853/en
    """
    global __model

    val = __model.scintillation_attenuation_synthesis(Ns, f_c, Ts)
    return val * u.dB


def integrated_water_vapour_synthesis(lat, lon, Ns, Ts=1, n=None):
    """ The time series synthesis method generates a time series that
    reproduces the spectral characteristics and the distribution of water
    vapour content.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    Ns : int
        Number of samples
    Ts : int
        Time step between consecutive samples (seconds)
    n : list, np.array, optional
        Additive White Gaussian Noise used as input for the

    Returns
    -------
    L: numpy.ndarray
        Synthesized water vapour content time series (kg/m2)


    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.1853/en
    """
    global __model

    lon = np.mod(lon, 360)
    val = __model.integrated_water_vapour_synthesis(lat, lon, Ns, Ts, n)
    return val * u.kg / u.m**2


def cloud_liquid_water_synthesis(lat, lon, Ns, Ts=1, n=None):
    """ The time series synthesis method generates a time series that
    reproduces the spectral characteristics, rate of change and duration
    statistics of cloud liquid content events.

    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    Ns : int
        Number of samples
    Ts : int
        Time step between consecutive samples (seconds)
    n : list, np.array, optional
        Additive White Gaussian Noise used as input for the

    Returns
    -------
    V: numpy.ndarray
        Synthesized cloud liquid water time series (mm)


    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.1853/en
    """
    global __model

    lon = np.mod(lon, 360)
    val = __model.cloud_liquid_water_synthesis(lat, lon, Ns, Ts, n)
    return val * u.mm


def total_attenuation_synthesis(lat, lon, f, el, p, D, Ns, Ts=1, hs=None,
                                tau=45, eta=0.65,  rho=None, H=None, P=None,
                                hL=1000, return_contributions=False):
    """ The time series synthesis method generates a time series that
    reproduces the spectral characteristics, rate of change and duration
    statistics of the total atmospheric attenuation events.

    The time series is obtained considering the contributions of gaseous,
    cloud, rain, and scintillation attenuation.

    Parameters
    ----------
    lat : number
        Latitudes of the receiver points
    lon : number
        Longitudes of the receiver points
    f : number or Quantity
        Frequency (GHz)
    el : number
        Elevation angle (degrees)
    p : number
        Percentage of the time the rain attenuation value is exceeded.
    D: number or Quantity
        Physical diameter of the earth-station antenna (m)
    Ns : int
        Number of samples
    Ts : int
        Time step between consecutive samples (seconds)
    tau : number, optional
        Polarization tilt angle relative to the horizontal (degrees)
        (tau = 45 deg for circular polarization). Default value is 45
    hs : number, sequence, or numpy.ndarray, optional
        Heigh above mean sea level of the earth station (km). If local data for
        the earth station height above mean sea level is not available, an
        estimate is obtained from the maps of topographic altitude
        given in Recommendation ITU-R P.1511.
    eta: number, optional
        Antenna efficiency. Default value 0.5 (conservative estimate)
    rho : number or Quantity, optional
        Water vapor density (g/m3). If not provided, an estimate is obtained
        from Recommendation Recommendation ITU-R P.836.
    H: number, sequence, or numpy.ndarray, optional
        Average surface relative humidity (%) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    P: number, sequence, or numpy.ndarray, optional
        Average surface pressure (hPa) at the site. If None, uses the
        ITU-R P.453 to estimate the wet term of the radio refractivity.
    hL : number, optional
        Height of the turbulent layer (m). Default value 1000 m
    return_contributions: bool, optional
        Determines whether individual contributions from gases, rain, clouds
        and scintillation are returned in addition ot the total attenuation
        (True), or just the total atmospheric attenuation (False).
        Default is False

    Returns
    ---------
    A : Quantity
        Synthesized total atmospheric attenuation time series (dB)

    Ag, Ac, Ar, As, A : tuple
        Synthesized Gaseous, Cloud, Rain, Scintillation contributions to total
        attenuation time series, and synthesized total attenuation time seires
        (dB).

    References
    ----------
    [1] Characteristics of precipitation for propagation modelling
    https://www.itu.int/rec/R-REC-P.1853/en
    """
    global __model

    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(el, u.deg, 'Elevation angle')
    D = prepare_quantity(D, u.m, 'Antenna diameter')
    hs = prepare_quantity(
        hs, u.km, 'Heigh above mean sea level of the earth station')
    eta = prepare_quantity(eta, u.one, 'Antenna efficiency')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapor density')
    H = prepare_quantity(H, u.percent, 'Average surface relative humidity')
    P = prepare_quantity(P, u.hPa, 'Average surface pressure')
    hL = prepare_quantity(hL, u.m, 'Height of the turbulent layer')

    val = __model.total_attenuation_synthesis(
            lat, lon, f, el, p, D, Ns, Ts=Ts, tau=tau, hs=hs, eta=eta, rho=rho,
            H=H, P=P, hL=hL, return_contributions=return_contributions)
    if return_contributions:
        return tuple([v * u.dB for v in val])
    else:
        return val * u.dB



# -*- coding: utf-8 -*-
"""``itur.utils`` is a utilities library for ITU-Rpy.

This utility library for ITU-Rpy contains methods to:
    * Load data and build an interpolator object.
    * Prepare the input and output arrays, and handle unit transformations.
    * Compute distances and elevation angles between two points on Earth and
      or space.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import numbers
import numpy as np

from pyproj import Geod
from astropy import units as u


# Set the basepath for the module and the basepath for the data
dir_path = os.path.dirname(os.path.realpath(__file__))
dataset_dir = os.path.join(dir_path, 'data/')


# Define numeric types including numpy types
__NUMERIC_TYPES__ = [numbers.Number, int, float, complex,
                     np.float, np.float16, np.float32, np.float64,
                     np.int, np.int8, np.int16, np.int32, np.int64]

# Define the geodetic system using the WSG-84 ellipsoid
__wgs84_geod__ = Geod(ellps='WGS84')


def load_data_interpolator(path_lat, path_lon, path_data, interp_fcn,
                           flip_ud=True):
    """Load a lat-lon tabulated dataset and build an interpolator.

    Parameters
    ----------
    path_lat : string
        Path for the file containing the latitude values
    path_lon : string
        Path for the file containing the longitude values
    path_data : string
        Path for the file containing the data values
    interp_fcn : string
        The interpolation function to be used
    flip_ud : boolean
        Whether to flip the latitude and data arrays along the first axis. This
        is an artifact of the format that the ITU uses to encode its data,
        which is inconsistent across recommendations (in some recommendations,
        latitude are sorted in ascending order, in others they are sorted in
        descending order).

    Returns
    -------
    interp: interp_fcn
        An interpolator that given a latitude-longitude pair, returns the
        data value
    """
    vals = load_data(os.path.join(dataset_dir, path_data))
    lats = load_data(os.path.join(dataset_dir, path_lat))
    lons = load_data(os.path.join(dataset_dir, path_lon))
    if flip_ud:
        return interp_fcn(np.flipud(lats), lons, np.flipud(vals))
    else:
        return interp_fcn(lats, lons, vals)


def load_data(path, is_text=False, **kwargs):
    """Load data files from `./itur/data/`.

    Loads data from a comma-separated values file. The contents of the file
    can be numeric or text-based.

    Parameters
    ----------
    path : string
        Path of the data to load
    is_text : bool
        Indicates whether the data is text (`True`) or numerical (`False`).
        Default value is `False`.

    Returns
    -------
    data: numpy.ndarray
        Numpy-array with the data. Numerical data is returned as a float
    """
    # TODO: Change method to allow for h5df data too
    if not os.path.isfile(path):
        raise RuntimeError('The path provided is not a file - {0}'
                           .format(path))

    _, file_extension = os.path.splitext(path)

    if file_extension == '.npz':
        data = np.load(path)['arr_0']
    elif file_extension == '.npy':
        data = np.load(path)
    elif file_extension == '.txt':
        if is_text:
            data = np.loadtxt(path, dtype=np.string_, delimiter=',', **kwargs)
        else:
            data = np.genfromtxt(path, dtype=float, delimiter=',', **kwargs)

    return data


def prepare_input_array(input_array):
    """Format an array to be a 2-D numpy-array.

    If the contents of `input_array` are 0-D or 1-D, it converts is to an
    array with at least two dimensions.

    Parameters
    ----------
    input_array : numpy.ndarray, sequence, or number
        The input value. It can be a scalar, 1-D array, or 2-D array.

    Returns
    -------
    output_array : numpy.ndarray
        An 2-D numpy array with the input values

    """
    if input_array is None:
        return None

    return np.atleast_2d(input_array)


def prepare_output_array(output_array, type_input=None):
    """Format the output to have the same shape and type as the input.

    This function is a generic wrapper to format the output of a function
    to have the same type as the input. ITU-Rpy makes extensive use of numpy
    arrays, but uses this function to return outputs having the same type
    that was provided in the input of the function.
    """
    # First, differentiate between the units and the value of the output_array
    # since the rest of the funcion is mainly focused on casting the value
    # of the output_array to the type in type_input
    if isinstance(output_array, u.Quantity):
        value = output_array.value
        unit = output_array.unit
    else:
        value = output_array
        unit = None

    # Squeeze output array to remove singleton dimensions
    if isinstance(value, np.ndarray) or isinstance(value, list):
        value = np.array(value).squeeze()

    type_output = type(output_array)
    # First, cast the output_array to the same type of the input
    # Check if the output array is a 0-D number and cast it to a float
    if (type_input in __NUMERIC_TYPES__ and
        (type_output in __NUMERIC_TYPES__) or
        ((isinstance(output_array, np.ndarray) and output_array.size == 1) or
         (not type_output not in __NUMERIC_TYPES__ and
          len(output_array) == 1))):
        value = float(value)

    # Check if the input array was a list and conver appropriately
    elif type_input is list:
        if isinstance(value, np.ndarray):
            value = value.tolist()
        else:
            value = list(value)

    # Otherwise, we assume that the value already has the required type
    else:
        value = value

    # Add the units of the
    if unit is not None:
        return value * unit
    else:
        return value


def prepare_quantity(value, units=None, name_val=None):
    """Convert the input to the required units.

    The function verifies that the input has the right units and converts
    it to the desired units. For example, if a value is introduced in km
    but posterior frequencies require this value to be in meters, this
    function would be called with `units=u.m`

    Parameters
    ----------
    value : astropy.units.Quantity, number, sequence, or np.ndarry
        The input value
    units : astropy.units
        Desired units of the output
    name_val : string
        Name of the variable (for debugging purposes)

    Returns
    -------
    q : numpy.ndarray
        An numpy array with the values converted to the desired units.
    """
    if value is None:
        return None

    # If the units of the value are a temperature
    if isinstance(value, u.Quantity):
        if units in [u.K, u.deg_C, u.Kelvin, u.Celsius, u.imperial.deg_F]:
            return value.to(units, equivalencies=u.temperature()).value
        else:
            return value.to(units).value
    # Process numbers
    elif isinstance(value, numbers.Number) and units is not None:
        return value
    # Process arrays and tuples
    elif isinstance(value, np.ndarray) and units is not None:
        return value
    elif isinstance(value, list) and units is not None:
        return np.array([prepare_quantity(v, units, name_val) for v in value])
    elif isinstance(value, tuple) and units is not None:
        return np.array([prepare_quantity(v, units, name_val) for v in value])
    else:
        raise ValueError('%s has not the correct format. It must be a value,'
                         'sequence, array, or a Quantity with %s units' %
                         (name_val, str(units)))


def compute_distance_earth_to_earth(lat_p, lon_p, lat_grid, lon_grid,
                                    method=None):
    """
    Compute the distance between a point and a matrix of (lat, lons).

    If the number of elements in `lat_grid` is smaller than 100,000, uses the
    WGS84 method, otherwise, uses the Haversine formula.

    Parameters
    ----------
    lat_p : number
        Latitude projection of the point P (degrees)
    lon_p : number
        Longitude projection of the point P (degrees)
    lat_grid : number, sequence of np.ndarray
        Grid of latitude points to which compute the distance (degrees)
    lon_grid : number, sequence of np.ndarray
        Grid of longitude points to which compute the distance (degrees)

    Returns
    -------
    d : numpy.ndarray
        Distance between the point P and each point in (lat_grid, lon_grid)
        (km)

    """
    if ((method == 'WGS84' and not(method is not None)) or
        (type(lat_p) in __NUMERIC_TYPES__) or
        (type(lat_grid) in __NUMERIC_TYPES__) or
        (len(lat_grid) < 10000) or
        (isinstance(lat_grid, np.ndarray) and lat_grid.size < 1e5)):
            return compute_distance_earth_to_earth_wgs84(
                    lat_p, lon_p, lat_grid, lon_grid)
    else:
        return compute_distance_earth_to_earth_haversine(
                lat_p, lon_p, lat_grid, lon_grid)


def compute_distance_earth_to_earth_wgs84(lat_p, lon_p, lat_grid, lon_grid):
    """Compute the distance between points using the WGS84 inverse method.

    Compute the distance between a point (P) in (`lat_p`, `lon_p`) and a matrix
    of latitude and longitudes (`lat_grid`, `lon_grid`) using the WGS84 inverse
    method.

    Parameters
    ----------
    lat_p : number
        Latitude projection of the point P (degrees)
    lon_p : number
        Longitude projection of the point P (degrees)
    lat_grid : number, sequence of np.ndarray
        Grid of latitude points to which compute the distance (degrees)
    lon_grid : number, sequence of np.ndarray
        Grid of longitude points to which compute the distance (degrees)

    Returns
    -------
    d : numpy.ndarray
        Distance between the point P and each point in (lat_grid, lon_grid)
        (km)

    """
    lat_p = lat_p * np.ones_like(lat_grid)
    lon_p = lon_p * np.ones_like(lon_grid)
    _a, _b, d = __wgs84_geod__.inv(lon_p, lat_p, lon_grid, lat_grid)
    return d / 1e3


def compute_distance_earth_to_earth_haversine(lat_p, lon_p,
                                              lat_grid, lon_grid):
    """Compute the distance between points using the Haversine formula.

    Compute the distance between a point (P) in (`lat_s`, `lon_s`) and a matrix
    of latitude and longitudes (`lat_grid`, `lon_grid`) using the Haversine
    formula.

    Parameters
    ----------
    lat_p : number
        Latitude projection of the point P (degrees)
    lon_p : number
        Longitude projection of the point P (degrees)
    lat_grid : number, sequence of np.ndarray
        Grid of latitude points to which compute the distance (degrees)
    lon_grid : number, sequence of np.ndarray
        Grid of longitude points to which compute the distance (degrees)

    Returns
    -------
    d : numpy.ndarray
        Distance between the point P and each point in (lat_grid, lon_grid)
        (km)

    References
    ----------
    This is based on the Haversine formula
    """
    RE = 6371.0  # Radius of the Earth, km

    lat1 = np.deg2rad(lat_grid)
    lat2 = np.deg2rad(lat_p)
    lon1 = np.deg2rad(lon_grid)
    lon2 = np.deg2rad(lon_p)

    dlat = lat2 - lat1
    dlon = lon2 - lon1

    # Compute the distance
    a = np.clip((np.sin(dlat / 2.0))**2 + np.cos(lat1) * np.cos(lat2) *
                (np.sin(dlon / 2))**2, -1, 1)
    c = 2 * np.arcsin(np.sqrt(a))
    d = RE * c
    return d


def regular_lat_lon_grid(resolution_lat=1, resolution_lon=1, lon_start_0=False,
                         lat_min=-90, lat_max=90, lon_min=-180, lon_max=180):
    """
    Build regular latitude and longitude matrices.

    Builds a latitude and longitude coordinate matrix with resolution
    `resolution_lat`, `resolution_lon`.

    Parameters
    ----------
    resolution_lat: number
        Resolution for the latitude axis (deg)
    resolution_lon: number
        Resolution for the longitude axis (deg)
    lon_start_0: boolean
        Indicates whether the longitude is indexed using a 0 - 360 scale (True)
        or using -180 - 180 scale (False). Default value is False

    Returns
    -------
    lat: numpy.ndarray
        Grid of coordinates of the latitude point
    lon: numpy.ndarray
        Grid of coordinates of the longitude point
    """
    if lon_start_0:
        lon, lat = np.meshgrid(np.arange(lon_min + 180.0, lon_max + 180.0,
                                         resolution_lon),
                               np.arange(lat_max, lat_min, - resolution_lat))
    else:
        lon, lat = np.meshgrid(np.arange(lon_min, lon_max, resolution_lon),
                               np.arange(lat_max, lat_min, - resolution_lat))

    return lat, lon


def elevation_angle(h, lat_s, lon_s, lat_grid, lon_grid):
    """
    Compute the elevation angle between a satellite and a point on Earth.

    Compute the elevation angle between a satellite located in an orbit
    at height h and located above coordinates (`lat_s`, `lon_s`) and a matrix
    of latitude and longitudes (`lat_grid`, `lon_grid`).

    Parameters
    ----------
    h : float
        Orbital altitude of the satellite (km)
    lat_s : float
        Latitude of the projection of the satellite (degrees)
    lon_s : float
        Longitude of the projection of the satellite (degrees)
    lat_grid :  number, sequence of np.ndarray
        Grid of latitude points to which compute the elevation angle (degrees)
    lon_grid :  number, sequence of np.ndarray
        Grid of longitude points to which compute the elevation angle (degrees)

    Returns
    -------
    elevation : numpy.ndarray
        Elevation angle between the satellite and each point in
        (lat_grid, lon_grid) (degrees)

    References
    ----------
    [1] http://www.propagation.gatech.edu/ECE6390/notes/ASD5.pdf - Slides 3, 4
    """
    h = prepare_quantity(h, u.km, name_val='Orbital altitude of the satellite')

    RE = 6371.0     # Radius of the Earth (km)
    rs = RE + h

    # Transform latitude_longitude values to radians
    lat1 = np.deg2rad(lat_grid)
    lat2 = np.deg2rad(lat_s)
    lon1 = np.deg2rad(lon_grid)
    lon2 = np.deg2rad(lon_s)

    # Compute the elevation angle as described in
    gamma = np.arccos(
        np.clip(np.sin(lat2) * np.sin(lat1) +
                np.cos(lat1) * np.cos(lat2) * np.cos(lon2 - lon1), -1, 1))
    elevation = np.arccos(np.sin(gamma) /
                          np.sqrt(1 + (RE / rs)**2 -
                                  2 * (RE / rs) * np.cos(gamma)))  # In radians

    return np.rad2deg(elevation)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import warnings


import numpy as np
from astropy import units as u

from itur.models.itu453 import radio_refractive_index
from itur.models.itu835 import (standard_pressure, standard_temperature,
                                standard_water_vapour_density)
from itur.models.itu836 import total_water_vapour_content
from itur.models.itu1511 import topographic_altitude
from itur.utils import (prepare_quantity, prepare_output_array,
                        prepare_input_array, load_data, dataset_dir)


def __gamma0_exact__(self, f, p, rho, T):
    # T in Kelvin
    # e : water vapour partial pressure in hPa (total barometric pressure
    # ptot = p + e)
    theta = 300 / T
    e = rho * T / 216.7

    f_ox = self.f_ox

    D_f_ox = self.a3 * 1e-4 * (p * (theta ** (0.8 - self.a4)) +
                               1.1 * e * theta)

    D_f_ox = np.sqrt(D_f_ox**2 + 2.25e-6)

    delta_ox = (self.a5 + self.a6 * theta) * 1e-4 * (p + e) * theta**0.8

    F_i_ox = f / f_ox * ((D_f_ox - delta_ox * (f_ox - f)) /
                         ((f_ox - f) ** 2 + D_f_ox ** 2) +
                         (D_f_ox - delta_ox * (f_ox + f)) /
                         ((f_ox + f) ** 2 + D_f_ox ** 2))

    Si_ox = self.a1 * 1e-7 * p * theta**3 * np.exp(self.a2 * (1 - theta))

    N_pp_ox = Si_ox * F_i_ox

    d = 5.6e-4 * (p + e) * theta**0.8

    N_d_pp = f * p * theta**2 * \
        (6.14e-5 / (d * (1 + (f / d)**2)) +
         1.4e-12 * p * theta**1.5 / (1 + 1.9e-5 * f**1.5))

    N_pp = N_pp_ox.sum() + N_d_pp

    gamma = 0.1820 * f * N_pp           # Eq. 1 [dB/km]
    return gamma


def __gammaw_exact__(self, f, p, rho, T):
    # T in Kelvin
    # e : water vapour partial pressure in hPa (total barometric pressure
    # ptot = p + e)
    theta = 300 / T
    e = rho * T / 216.7

    f_wv = self.f_wv

    D_f_wv = self.b3 * 1e-4 * (p * theta ** self.b4 +
                               self.b5 * e * theta ** self.b6)

    D_f_wv = 0.535 * D_f_wv + \
        np.sqrt(0.217 * D_f_wv**2 + 2.1316e-12 * f_wv**2 / theta)

    F_i_wv = f / f_wv * ((D_f_wv) / ((f_wv - f)**2 + D_f_wv**2) +
                         (D_f_wv) / ((f_wv + f)**2 + D_f_wv**2))

    Si_wv = self.b1 * 1e-1 * e * theta**3.5 * np.exp(self.b2 * (1 - theta))

    N_pp_wv = Si_wv * F_i_wv

    N_pp = N_pp_wv.sum()

    gamma = 0.1820 * f * N_pp           # Eq. 1 [dB/km]
    return gamma


class __ITU676__():
    """Attenuation by atmospheric gases.

    Available versions include:
       * P.676-9 (02/12) (Superseded)
       * P.676-10 (09/13) (Superseded)
       * P.676-11 (09/16) (Superseded)
       * P.676-11 (08/19) (Current version)

    Not available versions:
       * P.676-1 (03/92) (Superseded)
       * P.676-2 (10/95) (Superseded)
       * P.676-3 (08/97) (Superseded)
       * P.676-4 (10/99) (Superseded)
       * P.676-5 (02/01) (Superseded)
       * P.676-6 (03/05) (Superseded)
       * P.676-7 (02/07) (Superseded)
       * P.676-8 (10/09) (Superseded)

    """
    # This is an abstract class that contains an instance to a version of the
    # ITU-R P.676 recommendation.

    def __init__(self, version=12):
        if version == 12:
            self.instance = _ITU676_12_()
        elif version == 11:
            self.instance = _ITU676_11_()
        elif version == 10:
            self.instance = _ITU676_10_()
        elif version == 9:
            self.instance = _ITU676_9_()
#        elif version == 8:
#            self.instance = _ITU676_8()
#        elif version == 7:
#            self.instance = _ITU676_7()
#        elif version == 6:
#            self.instance = _ITU676_6()
#        elif version == 5:
#            self.instance = _ITU676_5()
#        elif version == 4:
#            self.instance = _ITU676_4()
#        elif version == 3:
#            self.instance = _ITU676_3()
#        elif version == 2:
#            self.instance = _ITU676_2()
#        elif version == 1:
#            self.instance = _ITU676_1()
        else:
            raise ValueError(
                'Version {0} is not implemented for the ITU-R P.676 model.'
                .format(version))

    @property
    def __version__(self):
        return self.instance.__version__

    def gaseous_attenuation_terrestrial_path(self, r, f, el, rho, P, T, mode):
        # Abstract method to compute the gaseous attenuation over a slant path
        fcn = np.vectorize(self.instance.gaseous_attenuation_terrestrial_path)
        return fcn(r, f, el, rho, P, T, mode)

    def gaseous_attenuation_inclined_path(self, f, el, rho, P, T, h1, h2, mode):
        # Abstract method to compute the gaseous attenuation over an inclined path
        fcn = np.vectorize(self.instance.gaseous_attenuation_inclined_path)
        return fcn(f, el, rho, P, T, h1, h2, mode)

    def gaseous_attenuation_slant_path(self, f, el, rho, P, T, V_t, h, mode):
        # Abstract method to compute the gaseous attenuation over a slant path
        fcn = np.vectorize(self.instance.gaseous_attenuation_slant_path)
        return fcn(f, el, rho, P, T, V_t, h, mode)

    def slant_inclined_path_equivalent_height(self, f, p):
        fcn = np.vectorize(self.instance.slant_inclined_path_equivalent_height,
                           excluded=[0], otypes=[np.ndarray])
        return np.array(fcn(f, p).tolist())

    def zenit_water_vapour_attenuation(
            self, lat, lon, p, f, V_t=None, h=None):
        # Abstract method to compute the water vapour attenuation over the
        # slant path
        fcn = np.vectorize(self.instance.zenit_water_vapour_attenuation,
                           excluded=[0, 1, 4, 5], otypes=[np.ndarray])
        return np.array(fcn(lat, lon, p, f, V_t, h).tolist())

    def gamma_exact(self, f, p, rho, t):
        # Abstract method to compute the specific attenuation using the
        # line-by-line method
        fcn = np.vectorize(self.instance.gamma_exact)
        return fcn(f, p, rho, t)

    def gammaw_exact(self, f, p, rho, t):
        # Abstract method to compute the specific attenuation due to water
        # vapour
        fcn = np.vectorize(self.instance.gammaw_exact)
        return fcn(f, p, rho, t)

    def gamma0_exact(self, f, p, rho, t):
        # Abstract method to compute the specific attenuation due to dry
        # atmoshere
        fcn = np.vectorize(self.instance.gamma0_exact)
        return fcn(f, p, rho, t)

    def gammaw_approx(self, f, p, rho, t):
        # Abstract method to compute the specific attenuation due to water
        # vapour
        fcn = np.vectorize(self.instance.gammaw_approx)
        return fcn(f, p, rho, t)

    def gamma0_approx(self, f, p, rho, t):
        # Abstract method to compute the specific attenuation due to dry
        # atmoshere
        fcn = np.vectorize(self.instance.gamma0_approx)
        return fcn(f, p, rho, t)


class _ITU676_12_():

    tmp = load_data(os.path.join(dataset_dir, '676/v12_lines_oxygen.txt'),
                    skip_header=1)
    f_ox = tmp[:, 0]
    a1 = tmp[:, 1]
    a2 = tmp[:, 2]
    a3 = tmp[:, 3]
    a4 = tmp[:, 4]
    a5 = tmp[:, 5]
    a6 = tmp[:, 6]

    tmp = load_data(os.path.join(dataset_dir,
                                 '676//v12_lines_water_vapour.txt'),
                    skip_header=1)
    f_wv = tmp[:, 0]
    b1 = tmp[:, 1]
    b2 = tmp[:, 2]
    b3 = tmp[:, 3]
    b4 = tmp[:, 4]
    b5 = tmp[:, 5]
    b6 = tmp[:, 6]

    # Coefficients in table 3
    t2_coeffs = [(0.1597, 118.750334),
                 (0.1066, 368.498246),
                 (0.1325, 424.763020),
                 (0.1242, 487.249273),
                 (0.0938, 715.392902),
                 (0.1448, 773.839490),
                 (0.1374, 834.145546)]

    # Coefficients in table 4
    hw_coeffs = [(22.23508, 1.52, 2.56),
                 (183.310087, 7.62, 10.2),
                 (325.152888, 1.56, 2.7),
                 (380.197353, 4.15, 5.7),
                 (439.150807, 0.2, 0.91),
                 (448.001085, 1.63, 2.46),
                 (474.689092, 0.76, 2.22),
                 (488.490108, 0.26, 2.49),
                 (556.935985, 7.81, 10),
                 (620.70087, 1.25, 2.35),
                 (752.033113, 16.2, 20),
                 (916.171582, 1.47, 2.58),
                 (970.315022, 1.36, 2.44),
                 (987.926764, 1.6, 1.86)]

    def __init__(self):
        self.__version__ = 12
        self.year = 2019
        self.month = 8
        self.link = 'https://www.itu.int/rec/R-REC-P.676-11-201712-S/en'

    def gammaw_approx(self, f, p, rho, T):
        warnings.warn(
                RuntimeWarning(
                    'Recommendation ITU-R P.676-12 does not have an explicit '
                    'method to approximate gamma_w. The exact method shall be '
                    'used instead.'))
        return self.gamma_exact(f, p, rho, T)

    def gamma0_approx(self, f, p, rho, T):
        warnings.warn(
                RuntimeWarning(
                    'Recommendation ITU-R P.676-12 does not have an explicit '
                    'method to approximate gamma_w. The exact method shall be '
                    'used instead.'))
        return self.gamma_exact(f, p, rho, T)

    @classmethod
    def gamma0_exact(self, f, p, rho, T):
        return __gamma0_exact__(self, f, p, rho, T)

    @classmethod
    def gammaw_exact(self, f, p, rho, T):
        return __gammaw_exact__(self, f, p, rho, T)

    @classmethod
    def gamma_exact(self, f, p, rho, T):
        return (self.gamma0_exact(f, p, rho, T) +
                self.gammaw_exact(f, p, rho, T))

    @classmethod
    def gaseous_attenuation_approximation(self, f, el, rho, P, T):
        """
        T goes in Kelvin
        """
        if np.any(f > 350):
            warnings.warn(
                RuntimeWarning(
                    'The approximated method to computes '
                    'the gaseous attenuation in recommendation ITU-P 676-11 '
                    'is only recommended for frequencies below 350GHz'))

        if np.any(5 > el) or np.any(np.mod(el, 90) < 5):
            warnings.warn(
                RuntimeWarning(
                    'The approximated method to compute '
                    'the gaseous attenuation in recommendation ITU-P 676-11 '
                    'is only recommended for elevation angles between'
                    '5 and 90 degrees'))

        # Water vapour attenuation (gammaw) computation as in Section 1 of
        # Annex 2 of [1]
        gamma0 = self.gamma0_exact(f, P, rho, T)
        gammaw = self.gammaw_exact(f, P, rho, T)

        return gamma0, gammaw

    @classmethod
    def slant_inclined_path_equivalent_height(self, f, P, rho, T):
        """
        """
        e = rho * T / 216.7
        rp = (P + e) / 1013.25

        # Eq. 31 - 34
        t1 = 5.1040 / (1 + 0.066 * rp**-2.3) * \
            np.exp(-((f - 59.7) / (2.87 + 12.4 * np.exp(-7.9 * rp)))**2)

        t2 = sum([(ci * np.exp(2.12 * rp)) /
                  ((f - fi)**2 + 0.025 * np.exp(2.2 * rp))
                  for ci, fi in self.t2_coeffs])

        t3 = 0.0114 * f / (1 + 0.14 * rp**-2.6) * \
            (15.02 * f**2 - 1353 * f + 5.333e4) / \
            (f**3 - 151.3 * f**2 + 9629 * f - 6803)

        A = 0.7832 + 0.00709 * (T - 273.15)

        # Eq. 30
        h0 = 6.1 * A / (1 + 0.17 * rp**-1.1) * (1 + t1 + t2 + t3)

        h0 = np.where(f < 70,
                      np.minimum(h0, 10.7 * rp**0.3),
                      h0)

        # Eq. 36 - 38
        A = 1.9298 - 0.04166 * (T - 273.15) + 0.0517 * rho
        B = 1.1674 - 0.00622 * (T - 273.15) + 0.0063 * rho
        sigmaw = 1.013 / (1 + np.exp(-8.6 * (rp - 0.57)))

        # Eq. 35 b
        hw = A + B * sum([(ai * sigmaw) / ((f - fi)**2 + bi * sigmaw)
                          for fi, ai, bi in self.hw_coeffs])
        return h0, hw

    @classmethod
    def gaseous_attenuation_terrestrial_path(
            self, r, f, el, rho, P, T, mode='approx'):
        """
        """
        if mode == 'approx':
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)
            return (gamma0 + gammaw) * r
        else:
            gamma = self.gamma_exact(f, P, rho, T)
            return gamma * r

    @classmethod
    def gaseous_attenuation_slant_path(self, f, el, rho, P, T, V_t=None,
                                       h=None, mode='approx'):
        """
        """
        if mode == 'approx':
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)

            h0, hw = self.slant_inclined_path_equivalent_height(f, P, rho, T)

            # Use the zenit water-vapour method if the values of V_t
            # and h are provided
            if V_t is not None and h is not None:
                Aw = self.zenit_water_vapour_attenuation(None, None, None,
                                                         f, V_t, h)
            else:
                Aw = gammaw * hw

            A0 = gamma0 * h0
            return (A0 + Aw) / np.sin(np.deg2rad(el))

        else:
            delta_h = 0.0001 * np.exp((np.arange(0, 923)) / 100)
            h_n = np.cumsum(delta_h)
            T_n = standard_temperature(h_n).to(u.K).value
            press_n = standard_pressure(h_n).value
            rho_n = standard_water_vapour_density(h_n, rho_0=rho).value

            e = rho * T / 216.7
            n_n = radio_refractive_index(press_n, e, T).value
            n_ratio = np.pad(n_n[1:], (0, 1), mode='edge') / n_n
            r_n = 6371 + h_n

            b = np.pi / 2 - np.deg2rad(el)
            Agas = 0
            for t, press, rho, r, delta, n_r in zip(
                    T_n, press_n, rho_n, r_n, delta_h, n_ratio):
                a = - r * np.cos(b) + 0.5 * np.sqrt(
                    4 * r**2 * np.cos(b)**2 + 8 * r * delta + 4 * delta**2)
                a_cos_arg = np.clip((-a**2 - 2 * r * delta - delta**2) /
                                    (2 * a * r + 2 * a * delta), -1, 1)
                alpha = np.pi - np.arccos(a_cos_arg)
                gamma = self.gamma_exact(f, press, rho, t)
                Agas += a * gamma
                b = np.arcsin(n_r * np.sin(alpha))

            return Agas

    @classmethod
    def gaseous_attenuation_inclined_path(
            self, f, el, rho, P, T, h1, h2, mode='approx'):
        """
        """
        if h1 > 10 or h2 > 10:
            raise ValueError(
                'Both the transmitter and the receiver must be at'
                'altitude of less than 10 km above the sea level.'
                'Current altitude Tx: %.2f km, Rx: %.2f km' % (h1, h2))

        if mode == 'approx':
            rho = rho * np.exp(h1 / 2)
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)
        else:
            gamma0 = self.gamma0_exact(f, P, rho, T)
            gammaw = 0

        e = rho * T / 216.7
        h0, hw = self.slant_inclined_path_equivalent_height(f, P + e)

        if 5 < el and el < 90:
            h0_p = h0 * (np.exp(-h1 / h0) - np.exp(-h2 / h0))
            hw_p = hw * (np.exp(-h1 / hw) - np.exp(-h2 / hw))
            return (gamma0 * h0_p + gammaw * hw_p) / np.sin(np.deg2rad(el))
        else:
            def F(x):
                return 1 / (0.661 * x + 0.339 * np.sqrt(x**2 + 5.51))

            el1 = el
            Re = 8500  # TODO: change to ITU-R P 834
            el2 = np.rad2deg(
                np.arccos(((Re + h1)/(Re + h2))*np.cos(np.deg2rad(el1))))

            def xi(eli, hi):
                return np.tan(np.deg2rad(eli)) * np.sqrt((Re + hi) / h0)

            def xi_p(eli, hi):
                return np.tan(np.deg2rad(eli)) * np.sqrt((Re + hi) / hw)

            def eq_33(h_num, h_den, el, x):
                return np.sqrt(Re + h_num) * F(x) * \
                    np.exp(-h_num / h_den) / np.cos(np.deg2rad(el))

            A = gamma0 * np.sqrt(h0) * (eq_33(h1, h0, el1, xi(el1, h1)) -
                                        eq_33(h2, h0, el2, xi(el2, h2))) +\
                gammaw * np.sqrt(hw) * (eq_33(h1, hw, el1, xi_p(el1, h1)) -
                                        eq_33(h2, hw, el2, xi_p(el2, h2)))
            return A

    @classmethod
    def zenit_water_vapour_attenuation(
            self, lat, lon, p, f, V_t=None, h=None):
        f_ref = 20.6        # [GHz]
        p_ref = 845         # [hPa]

        if h is None:
            h = topographic_altitude(lat, lon).value

        if V_t is None:
            V_t = total_water_vapour_content(lat, lon, p, h).value

        rho_ref = V_t / 2.38
        t_ref = 14 * np.log(0.22 * V_t / 2.38) + 3    # [Celsius]

        a = (0.2048 * np.exp(- ((f - 22.43)/3.097)**2) +
             0.2326 * np.exp(- ((f - 183.5)/4.096)**2) +
             0.2073 * np.exp(- ((f - 325)/3.651)**2) - 0.1113)

        b = 8.741e4 * np.exp(-0.587 * f) + 312.2 * f**(-2.38) + 0.723
        h = np.clip(h, 0, 4)

        gammaw_approx_vect = np.vectorize(self.gammaw_exact)

        Aw_term1 = (0.0176 * V_t *
                    gammaw_approx_vect(f, p_ref, rho_ref, t_ref + 273.15) /
                    gammaw_approx_vect(f_ref, p_ref, rho_ref, t_ref + 273.15))

        return np.where(f < 20, Aw_term1, Aw_term1 * (a * h ** b + 1))


class _ITU676_11_():

    tmp = load_data(os.path.join(dataset_dir, '676/v11_lines_oxygen.txt'),
                    skip_header=1)
    f_ox = tmp[:, 0]
    a1 = tmp[:, 1]
    a2 = tmp[:, 2]
    a3 = tmp[:, 3]
    a4 = tmp[:, 4]
    a5 = tmp[:, 5]
    a6 = tmp[:, 6]

    tmp = load_data(os.path.join(dataset_dir,
                                 '676//v11_lines_water_vapour.txt'),
                    skip_header=1)
    f_wv = tmp[:, 0]
    b1 = tmp[:, 1]
    b2 = tmp[:, 2]
    b3 = tmp[:, 3]
    b4 = tmp[:, 4]
    b5 = tmp[:, 5]
    b6 = tmp[:, 6]

    idx_approx = np.zeros_like(b1, dtype=bool).squeeze()
    asterisk_rows = [0, 3, 4, 5, 7, 12, 20, 24, 34]
    idx_approx[np.array(asterisk_rows)] = True

    def __init__(self):
        self.__version__ = 11
        self.year = 2017
        self.month = 12
        self.link = 'https://www.itu.int/rec/R-REC-P.676-11-201712-S/en'

    @classmethod
    def gammaw_approx(self, f, p, rho, T):
        # T in Kelvin
        # e : water vapour partial pressure in hPa (total barometric pressure
        # ptot = p + e)
        theta = 300 / T
        e = rho * T / 216.7

        f_wv = self.f_wv[self.idx_approx]
        b1 = self.b1[self.idx_approx]
        b2 = self.b2[self.idx_approx]
        b3 = self.b3[self.idx_approx]
        b4 = self.b4[self.idx_approx]
        b5 = self.b5[self.idx_approx]
        b6 = self.b6[self.idx_approx]

        D_f_wv = b3 * 1e-4 * (p * theta ** b4 +
                              b5 * e * theta ** b6)

        F_i_wv = f / f_wv * ((D_f_wv) / ((f_wv - f)**2 + D_f_wv**2) +
                             (D_f_wv) / ((f_wv + f)**2 + D_f_wv**2))

        Si_wv = b1 * 1e-1 * e * theta**3.5 * np.exp(b2 * (1 - theta))

        N_pp_wv = Si_wv * F_i_wv

        N_pp = N_pp_wv.sum()

        gamma = 0.1820 * f * N_pp           # Eq. 1 [dB/km]
        return gamma

    @classmethod
    def gamma0_approx(self, f, p, rho, T):
        # T in Kelvin
        # e : water vapour partial pressure in hPa (total barometric pressure
        # ptot = p + e)
        theta = 300 / T
        e = rho * T / 216.7

        f_ox = self.f_ox

        D_f_ox = self.a3 * 1e-4 * (p * (theta ** (0.8 - self.a4)) +
                                   1.1 * e * theta)

        delta_ox = (self.a5 + self.a6 * theta) * 1e-4 * (p + e) * theta**0.8

        F_i_ox = f / f_ox * ((D_f_ox - delta_ox * (f_ox - f)) /
                             ((f_ox - f) ** 2 + D_f_ox ** 2) +
                             (D_f_ox - delta_ox * (f_ox + f)) /
                             ((f_ox + f) ** 2 + D_f_ox ** 2))

        Si_ox = self.a1 * 1e-7 * p * theta**3 * np.exp(self.a2 * (1 - theta))

        N_pp_ox = Si_ox * F_i_ox

        d = 5.6e-4 * (p + e) * theta**0.8

        N_d_pp = f * p * theta**2 * \
            (6.14e-5 / (d * (1 + (f / d)**2)) +
             1.4e-12 * p * theta**1.5 / (1 + 1.9e-5 * f**1.5))

        N_pp = N_pp_ox.sum() + N_d_pp

        gamma = 0.1820 * f * N_pp           # Eq. 1 [dB/km]
        return gamma

    @classmethod
    def gamma0_exact(self, f, p, rho, T):
        return __gamma0_exact__(self, f, p, rho, T)

    @classmethod
    def gammaw_exact(self, f, p, rho, T):
        return __gammaw_exact__(self, f, p, rho, T)

    @classmethod
    def gamma_exact(self, f, p, rho, T):
        return (self.gamma0_exact(f, p, rho, T) +
                self.gammaw_exact(f, p, rho, T))

    @classmethod
    def gaseous_attenuation_approximation(self, f, el, rho, P, T):
        """
        T goes in Kelvin
        """
        if np.any(f > 350):
            warnings.warn(
                RuntimeWarning(
                    'The approximated method to computes '
                    'the gaseous attenuation in recommendation ITU-P 676-11 '
                    'is only recommended for frequencies below 350GHz'))

        if np.any(5 > el) or np.any(np.mod(el, 90) < 5):
            warnings.warn(
                RuntimeWarning(
                    'The approximated method to compute '
                    'the gaseous attenuation in recommendation ITU-P 676-11 '
                    'is only recommended for elevation angles between'
                    '5 and 90 degrees'))

        # Water vapour attenuation (gammaw) computation as in Section 1 of
        # Annex 2 of [1]
        gamma0 = self.gamma0_approx(f, P, rho, T)
        gammaw = self.gammaw_approx(f, P, rho, T)

        return gamma0, gammaw

    @classmethod
    def slant_inclined_path_equivalent_height(self, f, p):
        """
        """
        rp = p / 1013.25
        t1 = 4.64 / (1 + 0.066 * rp**-2.3) * \
            np.exp(- ((f - 59.7) / (2.87 + 12.4 * np.exp(-7.9 * rp)))**2)
        t2 = (0.14 * np.exp(2.12 * rp)) / \
            ((f - 118.75)**2 + 0.031 * np.exp(2.2 * rp))
        t3 = 0.0114 / (1 + 0.14 * rp**-2.6) * f * \
            (-0.0247 + 0.0001 * f + 1.61e-6 * f**2) / \
            (1 - 0.0169 * f + 4.1e-5 * f**2 + 3.2e-7 * f**3)

        h0 = 6.1 / (1 + 0.17 * rp**-1.1) * (1 + t1 + t2 + t3)

        h0 = np.where(f < 70,
                      np.minimum(h0, 10.7 * rp**0.3),
                      h0)

        sigmaw = 1.013 / (1 + np.exp(-8.6 * (rp - 0.57)))
        hw = 1.66 * (1 + (1.39 * sigmaw) / ((f - 22.235)**2 + 2.56 * sigmaw) +
                     (3.37 * sigmaw) / ((f - 183.31)**2 + 4.69 * sigmaw) +
                     (1.58 * sigmaw) / ((f - 325.1)**2 + 2.89 * sigmaw))

        return h0, hw

    @classmethod
    def gaseous_attenuation_terrestrial_path(
            self, r, f, el, rho, P, T, mode='approx'):
        """
        """
        if mode == 'approx':
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)
            return (gamma0 + gammaw) * r
        else:
            gamma = self.gamma_exact(f, P, rho, T)
            return gamma * r

    @classmethod
    def gaseous_attenuation_slant_path(self, f, el, rho, P, T, V_t=None,
                                       h=None, mode='approx'):
        """
        """
        if mode == 'approx':
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)

            e = rho * T / 216.7
            h0, hw = self.slant_inclined_path_equivalent_height(f, P + e)

            # Use the zenit water-vapour method if the values of V_t
            # and h are provided
            if V_t is not None and h is not None:
                Aw = self.zenit_water_vapour_attenuation(None, None, None,
                                                         f, V_t, h)
            else:
                Aw = gammaw * hw

            A0 = gamma0 * h0
            return (A0 + Aw) / np.sin(np.deg2rad(el))

        else:
            delta_h = 0.0001 * np.exp((np.arange(0, 923)) / 100)
            h_n = np.cumsum(delta_h)
            T_n = standard_temperature(h_n).to(u.K).value
            press_n = standard_pressure(h_n).value
            rho_n = standard_water_vapour_density(h_n, rho_0=rho).value

            e = rho * T / 216.7
            n_n = radio_refractive_index(press_n, e, T).value
            n_ratio = np.pad(n_n[1:], (0, 1), mode='edge') / n_n
            r_n = 6371 + h_n

            b = np.pi / 2 - np.deg2rad(el)
            Agas = 0
            for t, press, rho, r, delta, n_r in zip(
                    T_n, press_n, rho_n, r_n, delta_h, n_ratio):
                a = - r * np.cos(b) + 0.5 * np.sqrt(
                    4 * r**2 * np.cos(b)**2 + 8 * r * delta + 4 * delta**2)
                a_cos_arg = np.clip((-a**2 - 2 * r * delta - delta**2) /
                                    (2 * a * r + 2 * a * delta), -1, 1)
                alpha = np.pi - np.arccos(a_cos_arg)
                gamma = self.gamma_exact(f, press, rho, t)
                Agas += a * gamma
                b = np.arcsin(n_r * np.sin(alpha))

            return Agas

    @classmethod
    def gaseous_attenuation_inclined_path(
            self, f, el, rho, P, T, h1, h2, mode='approx'):
        """
        """
        if h1 > 10 or h2 > 10:
            raise ValueError(
                'Both the transmitter and the receiver must be at'
                'altitude of less than 10 km above the sea level.'
                'Current altitude Tx: %.2f km, Rx: %.2f km' % (h1, h2))

        if mode == 'approx':
            rho = rho * np.exp(h1 / 2)
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)
        else:
            gamma0 = self.gamma0_exact(f, P, rho, T)
            gammaw = 0

        e = rho * T / 216.7
        h0, hw = self.slant_inclined_path_equivalent_height(f, P + e)

        if 5 < el and el < 90:
            h0_p = h0 * (np.exp(-h1 / h0) - np.exp(-h2 / h0))
            hw_p = hw * (np.exp(-h1 / hw) - np.exp(-h2 / hw))
            return (gamma0 * h0_p + gammaw * hw_p) / np.sin(np.deg2rad(el))
        else:
            def F(x):
                return 1 / (0.661 * x + 0.339 * np.sqrt(x**2 + 5.51))

            el1 = el
            Re = 8500  # TODO: change to ITU-R P 834
            el2 = np.rad2deg(
                np.arccos(((Re + h1)/(Re + h2))*np.cos(np.deg2rad(el1))))

            def xi(eli, hi):
                return np.tan(np.deg2rad(eli)) * np.sqrt((Re + hi) / h0)

            def xi_p(eli, hi):
                return np.tan(np.deg2rad(eli)) * np.sqrt((Re + hi) / hw)

            def eq_33(h_num, h_den, el, x):
                return np.sqrt(Re + h_num) * F(x) * \
                    np.exp(-h_num / h_den) / np.cos(np.deg2rad(el))

            A = gamma0 * np.sqrt(h0) * (eq_33(h1, h0, el1, xi(el1, h1)) -
                                        eq_33(h2, h0, el2, xi(el2, h2))) +\
                gammaw * np.sqrt(hw) * (eq_33(h1, hw, el1, xi_p(el1, h1)) -
                                        eq_33(h2, hw, el2, xi_p(el2, h2)))
            return A

    @classmethod
    def zenit_water_vapour_attenuation(
            self, lat, lon, p, f, V_t=None, h=None):
        f_ref = 20.6        # [GHz]
        p_ref = 815         # [hPa]

        if h is None:
            h = topographic_altitude(lat, lon).value

        if V_t is None:
            V_t = total_water_vapour_content(lat, lon, p, h).value

        rho_ref = V_t / 3.67
        t_ref = 14 * np.log(0.22 * V_t / 3.67) + 3    # [Celsius]

        a = (0.2048 * np.exp(- ((f - 22.43)/3.097)**2) +
             0.2326 * np.exp(- ((f-183.5)/4.096)**2) +
             0.2073 * np.exp(- ((f-325)/3.651)**2) - 0.113)

        b = 8.741e4 * np.exp(-0.587 * f) + 312.2 * f**(-2.38) + 0.723
        h = np.minimum(h, 4)

        gammaw_approx_vect = np.vectorize(self.gammaw_approx)

        Aw_term1 = (0.0176 * V_t *
                    gammaw_approx_vect(f, p_ref, rho_ref, t_ref + 273.15) /
                    gammaw_approx_vect(f_ref, p_ref, rho_ref, t_ref + 273.15))

        return np.where(f < 20, Aw_term1, Aw_term1 * (a * h ** b + 1))


class _ITU676_10_():

    tmp = load_data(os.path.join(dataset_dir, '676/v10_lines_oxygen.txt'),
                    skip_header=1)
    f_ox = tmp[:, 0]
    a1 = tmp[:, 1]
    a2 = tmp[:, 2]
    a3 = tmp[:, 3]
    a4 = tmp[:, 4]
    a5 = tmp[:, 5]
    a6 = tmp[:, 6]

    tmp = load_data(os.path.join(dataset_dir,
                                 '676//v10_lines_water_vapour.txt'),
                    skip_header=1)
    f_wv = tmp[:, 0]
    b1 = tmp[:, 1]
    b2 = tmp[:, 2]
    b3 = tmp[:, 3]
    b4 = tmp[:, 4]
    b5 = tmp[:, 5]
    b6 = tmp[:, 6]

    def __init__(self):
        self.__version__ = 10
        self.year = 2013
        self.month = 9
        self.link = 'https://www.itu.int/rec/R-REC-P.676-10-201309-S/en'

    @classmethod
    def gammaw_approx(self, f, P, rho, T):
        rp = P / 1013
        rt = 288 / (T)
        eta1 = 0.955 * rp * rt**0.68 + 0.006 * rho
        eta2 = 0.735 * rp * rt**0.50 + 0.0353 * rt**4 * rho

        def g(f, fi): return 1 + ((f - fi) / (f + fi))**2
        gammaw = (
            (3.98 * eta1 * np.exp(2.23 * (1 - rt))) /
            ((f - 22.235) ** 2 + 9.42 * eta1 ** 2) * g(f, 22.0) +
            (11.96 * eta1 * np.exp(0.70 * (1 - rt))) /
            ((f - 183.310) ** 2 + 11.14 * eta1 ** 2) +
            (0.081 * eta1 * np.exp(6.44 * (1 - rt))) /
            ((f - 321.226) ** 2 + 6.29 * eta1 ** 2) +
            (3.660 * eta1 * np.exp(1.60 * (1 - rt))) /
            ((f - 325.153) ** 2 + 9.22 * eta1 ** 2) +
            (25.37 * eta1 * np.exp(1.09 * (1 - rt))) / ((f - 380.000) ** 2) +
            (17.40 * eta1 * np.exp(1.46 * (1 - rt))) / ((f - 448.000) ** 2) +
            (844.6 * eta1 * np.exp(0.17 * (1 - rt))) / ((f - 557.000) ** 2) *
            g(f, 557.0) + (290.0 * eta1 * np.exp(0.41 * (1 - rt))) /
            ((f - 752.000) ** 2) * g(f, 752.0) +
            (8.3328e4 * eta2 * np.exp(0.99 * (1 - rt))) /
            ((f - 1780.00) ** 2) *
            g(f, 1780.0)) * f ** 2 * rt ** 2.5 * rho * 1e-4
        return gammaw

    @classmethod
    def gamma0_approx(self, f, P, rho, T):
        rp = P / 1013.0
        rt = 288.0 / (T)

        def phi(rp, rt, a, b, c, d): return np.power(
            rp, a) * np.power(rt, b) * np.exp(c * (1 - rp) + d * (1 - rt))

        # Dry air attenuation (gamma0) computation as in Section 1 of Annex 2
        # of [1]
        delta = -0.00306 * phi(rp, rt, 3.211, -14.94, 1.583, -16.37)

        xi1 = phi(rp, rt, 0.0717, -1.8132, 0.0156, -1.6515)
        xi2 = phi(rp, rt, 0.5146, -4.6368, -0.1921, -5.7416)
        xi3 = phi(rp, rt, 0.3414, -6.5851, 0.2130, -8.5854)
        xi4 = phi(rp, rt, -0.0112, 0.0092, -0.1033, -0.0009)
        xi5 = phi(rp, rt, 0.2705, -2.7192, -0.3016, -4.1033)
        xi6 = phi(rp, rt, 0.2445, -5.9191, 0.0422, -8.0719)
        xi7 = phi(rp, rt, -0.1833, 6.5589, -0.2402, 6.131)

        gamma54 = 2.192 * phi(rp, rt, 1.8286, -1.9487, 0.4051, -2.8509)
        gamma58 = 12.59 * phi(rp, rt, 1.0045, 3.5610, 0.1588, 1.2834)
        gamma60 = 15.00 * phi(rp, rt, 0.9003, 4.1335, 0.0427, 1.6088)
        gamma62 = 14.28 * phi(rp, rt, 0.9886, 3.4176, 0.1827, 1.3429)
        gamma64 = 6.819 * phi(rp, rt, 1.4320, 0.6258, 0.3177, -0.5914)
        gamma66 = 1.908 * phi(rp, rt, 2.0717, -4.1404, 0.4910, -4.8718)

        def fcn_le_54():
            return (((7.2 * rt**2.8) / (f**2 + 0.34 * rp**2 * rt**1.6) +
                     (0.62 * xi3) / ((54 - f)**(1.16 * xi1) + 0.83 * xi2)) *
                    f**2 * rp**2 * 1e-3)

        def fcn_le_60():
            return (np.exp(np.log(gamma54) / 24.0 * (f - 58) * (f - 60) -
                           np.log(gamma58) / 8.0 * (f - 54) * (f - 60) +
                           np.log(gamma60) / 12.0 * (f - 54) * (f - 58)))

        def fcn_le_62():
            return (gamma60 + (gamma62 - gamma60) * (f - 60) / 2.0)

        def fcn_le_66():
            return (np.exp(np.log(gamma62) / 8.0 * (f - 64) * (f - 66) -
                           np.log(gamma64) / 4.0 * (f - 62) * (f - 66) +
                           np.log(gamma66) / 8.0 * (f - 62) * (f - 64)))

        def fcn_le_120():
            return ((3.02e-4 * rt**3.5 + (0.283 * rt**3.8) /
                     ((f - 118.75)**2 + 2.91 * rp**2 * rt**1.6) +
                     (0.502 * xi6 * (1 - 0.0163 * xi7 * (f - 66))) /
                     ((f - 66)**(1.4346 * xi4) + 1.15 * xi5)) *
                    f**2 * rp**2 * 1e-3)

        def fcn_rest():
            return (((3.02e-4) / (1 + 1.9e-5 * f**1.5) +
                     (0.283 * rt**0.3) / ((f - 118.75)**2 +
                                          2.91 * rp**2 * rt**1.6)) *
                    f**2 * rp**2 * rt**3.5 * 1e-3 + delta)

        gamma0 = \
            np.where(
                f <= 54, fcn_le_54(),
                np.where(
                    np.logical_and(54 < f, f <= 60), fcn_le_60(),
                    np.where(
                        np.logical_and(60 < f, f <= 62), fcn_le_62(),
                        np.where(
                            np.logical_and(62 < f, f <= 66), fcn_le_66(),
                            np.where(
                                np.logical_and(66 < f, f <= 120),
                                fcn_le_120(),
                                fcn_rest())))))

        return gamma0

    @classmethod
    def gamma0_exact(self, f, p, rho, T):
        return __gamma0_exact__(self, f, p, rho, T)

    @classmethod
    def gammaw_exact(self, f, p, rho, T):
        return __gammaw_exact__(self, f, p, rho, T)

    @classmethod
    def gamma_exact(self, f, p, rho, T):
        return (self.gamma0_exact(f, p, rho, T) +
                self.gammaw_exact(f, p, rho, T))

    @classmethod
    def gaseous_attenuation_approximation(self, f, el, rho, P, T):
        """
        T goes in Kelvin
        """
        if np.any(f > 350):
            warnings.warn(
                RuntimeWarning(
                    'The approximated method to computes '
                    'the gaseous attenuation in recommendation ITU-P 676-11 '
                    'is only recommended for frequencies below 350GHz'))

        if np.any(5 > el) or np.any(np.mod(el, 90) < 5):
            warnings.warn(
                RuntimeWarning(
                    'The approximated method to compute '
                    'the gaseous attenuation in recommendation ITU-P 676-11 '
                    'is only recommended for elevation angles between'
                    '5 and 90 degrees'))

        # Water vapour attenuation (gammaw) computation as in Section 1 of
        # Annex 2 of [1]
        gamma0 = self.gamma0_approx(f, P, rho, T)
        gammaw = self.gammaw_approx(f, P, rho, T)

        return gamma0, gammaw

    @classmethod
    def slant_inclined_path_equivalent_height(self, f, p):
        """
        """
        rp = p / 1013.0
        t1 = (4.64) / (1 + 0.066 * rp**-2.3) * \
            np.exp(- ((f - 59.7) / (2.87 + 12.4 * np.exp(-7.9 * rp)))**2)
        t2 = (0.14 * np.exp(2.21 * rp)) / \
            ((f - 118.75)**2 + 0.031 * np.exp(2.2 * rp))
        t3 = (0.0114) / (1 + 0.14 * rp**-2.6) * f * \
             (-0.0247 + 0.0001 * f + 1.61e-6 * f**2) / \
             (1 - 0.0169 * f + 4.1e-5 * f**2 + 3.2e-7 * f**3)

        h0 = (6.1) / (1 + 0.17 * rp**-1.1) * (1 + t1 + t2 + t3)

        h0 = np.where(f < 70,
                      np.minimum(h0, 10.7 * rp**0.3),
                      h0)

        sigmaw = (1.013) / (1 + np.exp(-8.6 * (rp - 0.57)))
        hw = 1.66 * (1 + (1.39 * sigmaw) / ((f - 22.235)**2 + 2.56 * sigmaw) +
                     (3.37 * sigmaw) / ((f - 183.31)**2 + 4.69 * sigmaw) +
                     (1.58 * sigmaw) / ((f - 325.1)**2 + 2.89 * sigmaw))

        return h0, hw

    @classmethod
    def gaseous_attenuation_terrestrial_path(
            self, r, f, el, rho, P, T, mode='approx'):
        """
        """
        if mode == 'approx':
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)
            return (gamma0 + gammaw) * r
        else:
            gamma = self.gamma_exact(f, P, rho, T)
            return gamma * r

    @classmethod
    def gaseous_attenuation_slant_path(self, f, el, rho, P, T, V_t=None,
                                       h=None, mode='approx'):
        """
        """
        if mode == 'approx':
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)
            e = rho * T / 216.7
            h0, hw = self.slant_inclined_path_equivalent_height(f, P + e)

            # Use the zenit water-vapour method if the values of V_t
            # and h are provided
            if V_t is not None and h is not None:
                Aw = self.zenit_water_vapour_attenuation(None, None, None,
                                                         f, V_t, h)
            else:
                Aw = gammaw * hw

            A0 = gamma0 * h0
            return (A0 + Aw) / np.sin(np.deg2rad(el))

        else:
            delta_h = 0.0001 * np.exp((np.arange(1, 923) - 1) / 100)
            h_n = np.cumsum(delta_h)
            T_n = standard_temperature(h_n).to(u.K).value
            press_n = standard_pressure(h_n).value
            rho_n = standard_water_vapour_density(h_n, rho_0=rho).value

            e = rho * T / 216.7
            n_n = radio_refractive_index(press_n, e, T).value
            n_ratio = np.pad(n_n[1:], (0, 1), mode='edge') / n_n
            r_n = 6371 + h_n

            b = np.pi / 2 - np.deg2rad(el)
            Agas = 0
            for t, press, rho, r, delta, n_r in zip(
                    T_n, press_n, rho_n, r_n, delta_h, n_ratio):
                a = - r * np.cos(b) + 0.5 * np.sqrt(
                    4 * r**2 * np.cos(b)**2 + 8 * r * delta + 4 * delta**2)
                a_cos_arg = np.clip((-a**2 - 2 * r * delta - delta**2) /
                                    (2 * a * r + 2 * a * delta), -1, 1)
                alpha = np.pi - np.arccos(a_cos_arg)
                gamma = self.gamma_exact(f, press, rho, t)
                Agas += a * gamma
                b = np.arcsin(n_r * np.sin(alpha))

            return Agas

    @classmethod
    def gaseous_attenuation_inclined_path(
            self, f, el, rho, P, T, h1, h2, mode='approx'):
        """
        """
        if h1 > 10 or h2 > 10:
            raise ValueError(
                'Both the transmitter and the receiver must be at'
                'altitude of less than 10 km above the sea level.'
                'Current altitude Tx: %.2f km, Rx: %.2f km' % (h1, h2))

        if mode == 'approx':
            rho = rho * np.exp(h1 / 2)
            gamma0, gammaw = self.gaseous_attenuation_approximation(
                f, el, rho, P, T)
        else:
            gamma0 = self.gamma_exact(f, P, rho, T)
            gammaw = 0

        e = rho * T / 216.7
        h0, hw = self.slant_inclined_path_equivalent_height(f, P + e)

        if 5 < el and el < 90:
            h0_p = h0 * (np.exp(-h1 / h0) - np.exp(-h2 / h0))
            hw_p = hw * (np.exp(-h1 / hw) - np.exp(-h2 / hw))
            return (gamma0 * h0_p + gammaw * hw_p) / np.sin(np.deg2rad(el))
        else:
            def F(x):
                return 1 / (0.661 * x + 0.339 * np.sqrt(x**2 + 5.51))

            el1 = el
            Re = 8500  # TODO: change to ITU-R P 834
            el2 = np.rad2deg(
                np.arccos(((Re + h1)/(Re + h2))*np.cos(np.deg2rad(el1))))

            def xi(eli, hi):
                return np.tan(np.deg2rad(eli)) * np.sqrt((Re + hi) / h0)

            def xi_p(eli, hi):
                return np.tan(np.deg2rad(eli)) * np.sqrt((Re + hi) / hw)

            def eq_33(h_num, h_den, el, x):
                return np.sqrt(Re + h_num) * F(x) * \
                    np.exp(-h_num / h_den) / np.cos(np.deg2rad(el))

            A = gamma0 * np.sqrt(h0) * (eq_33(h1, h0, el1, xi(el1, h1)) -
                                        eq_33(h2, h0, el2, xi(el2, h2))) +\
                gammaw * np.sqrt(hw) * (eq_33(h1, hw, el1, xi_p(el1, h1)) -
                                        eq_33(h2, hw, el2, xi_p(el2, h2)))
            return A

    @classmethod
    def zenit_water_vapour_attenuation(
            self, lat, lon, p, f, V_t=None, h=None):
        f_ref = 20.6        # [GHz]
        p_ref = 780         # [hPa]
        if V_t is None:
            V_t = total_water_vapour_content(lat, lon, p, h).value
        rho_ref = V_t / 4     # [g/m3]
        t_ref = 14 * np.log(0.22 * V_t / 4) + 3    # [Celsius]

        gammaw_approx_vect = np.vectorize(self.gammaw_approx)
        return (0.0173 * V_t *
                gammaw_approx_vect(f, p_ref, rho_ref, t_ref + 273) /
                gammaw_approx_vect(f_ref, p_ref, rho_ref, t_ref + 273))


class _ITU676_9_():

    tmp = load_data(os.path.join(dataset_dir, '676//v9_lines_oxygen.txt'),
                    skip_header=1)
    f_ox = tmp[:, 0]
    a1 = tmp[:, 1]
    a2 = tmp[:, 2]
    a3 = tmp[:, 3]
    a4 = tmp[:, 4]
    a5 = tmp[:, 5]
    a6 = tmp[:, 6]

    tmp = load_data(os.path.join(dataset_dir,
                                 '676//v9_lines_water_vapour.txt'),
                    skip_header=1)
    f_wv = tmp[:, 0]
    b1 = tmp[:, 1]
    b2 = tmp[:, 2]
    b3 = tmp[:, 3]
    b4 = tmp[:, 4]
    b5 = tmp[:, 5]
    b6 = tmp[:, 6]

    def __init__(self):
        self.__version__ = 9
        self.year = 2012
        self.month = 2
        self.link = 'https://www.itu.int/rec/R-REC-P.676-9-201202-S/en'

    # Recommendation ITU-P R.676-9 has most of the methods similar to those
    # in Recommendation ITU-P R.676-10.
    @staticmethod
    def gammaw_approx(*args, **kwargs):
        return _ITU676_10_.gammaw_approx(*args, **kwargs)

    @staticmethod
    def gamma0_approx(*args, **kwargs):
        return _ITU676_10_.gamma0_approx(*args, **kwargs)

    @staticmethod
    def gaseous_attenuation_inclined_path(*args, **kwargs):
        return _ITU676_10_.gaseous_attenuation_inclined_path(*args, **kwargs)

    @staticmethod
    def zenit_water_vapour_attenuation(*args, **kwargs):
        return _ITU676_10_.zenit_water_vapour_attenuation(*args, **kwargs)

    @staticmethod
    def gaseous_attenuation_approximation(*args, **kwargs):
        return _ITU676_10_.gaseous_attenuation_approximation(*args, **kwargs)

    @staticmethod
    def slant_inclined_path_equivalent_height(*args, **kwargs):
        return _ITU676_10_.slant_inclined_path_equivalent_height(*args,
                                                                 **kwargs)

    @staticmethod
    def gaseous_attenuation_terrestrial_path(*args, **kwargs):
        return _ITU676_10_.gaseous_attenuation_terrestrial_path(*args,
                                                                **kwargs)

    @staticmethod
    def gaseous_attenuation_slant_path(*args, **kwargs):
        return _ITU676_10_.gaseous_attenuation_slant_path(*args, **kwargs)

    def gamma0_exact(self, f, p, rho, T):
        return __gamma0_exact__(self, f, p, rho, T)

    def gammaw_exact(self, f, p, rho, T):
        return __gammaw_exact__(self, f, p, rho, T)

    def gamma_exact(self, f, p, rho, T):
        return (self.gamma0_exact(f, p, rho, T) +
                self.gammaw_exact(f, p, rho, T))

__model = __ITU676__()


def change_version(new_version):
    """
    Change the version of the ITU-R P.676 recommendation currently being used.

    This function changes the model used for the ITU-R P.676 recommendation
    to a different version.


    Parameters
    ----------
    new_version : int
        Number of the version to use.
        Valid values are:
          *  12: Activates recommendation ITU-R P.676-12 (08/19) (Current version)
          *  11: Activates recommendation ITU-R P.676-11 (09/16) (Superseded)
          *  10: Activates recommendation ITU-R P.676-10 (09/13) (Superseded)
          *  9: Activates recommendation ITU-R P.676-9 (02/12) (Superseded)

    """
    global __model
    __model = __ITU676__(new_version)


def get_version():
    """
    Obtain the version of the ITU-R P.676 recommendation currently being used.

    Returns
    -------
    version: int
       The version of the ITU-R P.530 recommendation being used.
    """
    return __model.__version__


def gaseous_attenuation_terrestrial_path(r, f, el, rho, P, T, mode):
    """
    Estimate the attenuation of atmospheric gases on terrestrial paths.
    This function operates in two modes, 'approx', and 'exact':
      * 'approx': a simplified approximate method to estimate gaseous attenuation
        that is applicable in the frequency range 1-350 GHz.
      * 'exact': an estimate of gaseous attenuation computed by summation of
        individual absorption lines that is valid for the frequency
        range 1-1,000 GHz


    Parameters
    ----------
    r : number or Quantity
        Path length (km)
    f : number or Quantity
        Frequency (GHz)
    el : sequence, number or Quantity
        Elevation angle (degrees)
    rho : number or Quantity
        Water vapor density (g/m**3)
    P : number or Quantity
        Atmospheric pressure (hPa)
    T : number or Quantity
        Absolute temperature (K)
    mode : string, optional
        Mode for the calculation. Valid values are 'approx', 'exact'. If
        'approx' Uses the method in Annex 2 of the recommendation (if any),
        else uses the method described in Section 1. Default, 'approx'


    Returns
    -------
    attenuation: Quantity
        Terrestrial path attenuation (dB)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    type_output = type(el)
    r = prepare_quantity(r, u.km, 'Path Length')
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(prepare_input_array(el), u.deg, 'Elevation angle')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapor density')
    P = prepare_quantity(P, u.hPa, 'Atospheric pressure')
    T = prepare_quantity(T, u.K, 'Temperature')
    val = __model.gaseous_attenuation_terrestrial_path(
            r, f, el, rho, P, T, mode)
    return prepare_output_array(val, type_output) * u.dB


def gaseous_attenuation_slant_path(f, el, rho, P, T, V_t=None, h=None,
                                   mode='approx'):
    """
    Estimate the attenuation of atmospheric gases on slant paths. This function
    operates in two modes, 'approx', and 'exact':
      * 'approx': a simplified approximate method to estimate gaseous attenuation
        that is applicable in the frequency range 1-350 GHz.
      * 'exact': an estimate of gaseous attenuation computed by summation of
        individual absorption lines that is valid for the frequency
        range 1-1,000 GHz


    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    el : sequence, number or Quantity
        Elevation angle (degrees)
    rho : number or Quantity
        Water vapor density (g/m3)
    P : number or Quantity
        Atmospheric pressure (hPa)
    T : number or Quantity
        Absolute temperature (K)
    V_t: number or Quantity (kg/m2)
        Integrated water vapour content from: a) local radiosonde or
        radiometric data or b) at the required percentage of time (kg/m2)
        obtained from the digital maps in Recommendation ITU-R P.836 (kg/m2).
        If None, use general method to compute the wet-component of the
        gaseous attenuation. If provided, 'h' must be also provided. Default
        is None.
    h : number, sequence, or numpy.ndarray
        Altitude of the receivers. If None, use the topographical altitude as
        described in recommendation ITU-R P.1511. If provided, 'V_t' needs to
        be also provided. Default is None.
    mode : string, optional
        Mode for the calculation. Valid values are 'approx', 'exact'. If
        'approx' Uses the method in Annex 2 of the recommendation (if any),
        else uses the method described in Section 1. Default, 'approx'


    Returns
    -------
    attenuation: Quantity
        Slant path attenuation (dB)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    type_output = type(el)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(prepare_input_array(el), u.deg, 'Elevation angle')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapor density')
    P = prepare_quantity(P, u.hPa, 'Atospheric pressure')
    T = prepare_quantity(T, u.K, 'Temperature')
    V_t = prepare_quantity(V_t, u.kg / u.m**2,
                           'Integrated water vapour content')
    h = prepare_quantity(h, u.km, 'Altitude')
    val = __model.gaseous_attenuation_slant_path(
            f, el, rho, P, T, V_t, h, mode)
    return prepare_output_array(val, type_output) * u.dB


def gaseous_attenuation_inclined_path(f, el, rho, P, T, h1, h2, mode='approx'):
    """
    Estimate the attenuation of atmospheric gases on inclined paths between two
    ground stations at heights h1 and h2. This function operates in two modes,
    'approx', and 'exact':
      * 'approx': a simplified approximate method to estimate gaseous attenuation
        that is applicable in the frequency range 1-350 GHz.
      * 'exact': an estimate of gaseous attenuation computed by summation of
        individual absorption lines that is valid for the frequency
        range 1-1,000 GHz


    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    el : sequence, number or Quantity
        Elevation angle (degrees)
    rho : number or Quantity
        Water vapor density (g/m3)
    P : number or Quantity
        Atmospheric pressure (hPa)
    T : number or Quantity
        Absolute temperature (K)
    h1 : number or Quantity
        Height of ground station 1 (km)
    h2 : number or Quantity
        Height of ground station 2 (km)
    mode : string, optional
        Mode for the calculation. Valid values are 'approx', 'exact'. If
        'approx' Uses the method in Annex 2 of the recommendation (if any),
        else uses the method described in Section 1. Default, 'approx'


    Returns
    -------
    attenuation: Quantity
        Inclined path attenuation (dB)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    f = prepare_quantity(f, u.GHz, 'Frequency')
    el = prepare_quantity(el, u.deg, 'Elevation angle')
    type_output = type(el)
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapor density')
    P = prepare_quantity(P, u.hPa, 'Atospheric pressure')
    T = prepare_quantity(T, u.K, 'Temperature')
    h1 = prepare_quantity(h1, u.km, 'Height of Ground Station 1')
    h2 = prepare_quantity(h2, u.km, 'Height of Ground Station 2')
    val = __model.gaseous_attenuation_inclined_path(
        f, el, rho, P, T, h1, h2, mode=mode)
    return prepare_output_array(val, type_output) * u.dB


def slant_inclined_path_equivalent_height(f, p):
    """ Computes the equivalent height to be used for oxygen and water vapour
    gaseous attenuation computations.

    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    p : number
        Percentage of the time the gaseous attenuation value is exceeded.

    Returns
    -------
    ho, hw : Quantity
        Equivalent height for oxygen and water vapour (m)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en

    """
    type_output = type(f)
    f = prepare_quantity(f, u.GHz, 'Frequency')

    val = __model.slant_inclined_path_equivalent_height(f, p)
    return prepare_output_array(val, type_output) * u.m


def zenit_water_vapour_attenuation(lat, lon, p, f, V_t=None, h=None):
    """
    An alternative method may be used to compute the slant path attenuation by
    water vapour, in cases where the integrated water vapour content along the
    path, ``V_t``, is known.


    Parameters
    ----------
    lat : number, sequence, or numpy.ndarray
        Latitudes of the receiver points
    lon : number, sequence, or numpy.ndarray
        Longitudes of the receiver points
    p : number
        Percentage of the time the zenit water vapour attenuation value is
        exceeded.
    f : number or Quantity
        Frequency (GHz)
    V_t : number or Quantity, optional
        Integrated water vapour content along the path (kg/m2 or mm).
        If not provided this value is estimated using Recommendation
        ITU-R P.836. Default value None
    h : number, sequence, or numpy.ndarray
        Altitude of the receivers. If None, use the topographical altitude as
        described in recommendation ITU-R P.1511


    Returns
    -------
    A_w : Quantity
        Water vapour attenuation along the slant path (dB)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    type_output = type(lat)
    lat = prepare_input_array(lat)
    lon = prepare_input_array(lon)
    lon = np.mod(lon, 360)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    V_t = prepare_quantity(
        V_t,
        u.kg / u.m**2,
        'Integrated water vapour content along the path')
    h = prepare_quantity(h, u.km, 'Altitude')
    val = __model.zenit_water_vapour_attenuation(
        lat, lon, p, f, V_t=V_t, h=h)
    return prepare_output_array(val, type_output) * u.dB


def gammaw_approx(f, P, rho, T):
    """
    Method to estimate the specific attenuation due to water vapour using the
    approximate method descibed in Annex 2.


    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    P : number or Quantity
        Atmospheric pressure (hPa)
    rho : number or Quantity
        Water vapor density (g/m3)
    T : number or Quantity
        Absolute temperature (K)


    Returns
    -------
    gamma_w : Quantity
        Water vapour specific attenuation (dB/km)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    type_output = type(f)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    P = prepare_quantity(P, u.hPa, 'Atmospheric pressure ')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapour density')
    T = prepare_quantity(T, u.K, 'Temperature')
    val = __model.gammaw_approx(f, P, rho, T)
    return prepare_output_array(val, type_output) * u.dB / u.km


def gamma0_approx(f, P, rho, T):
    """
    Method to estimate the specific attenuation due to dry atmosphere using the
    approximate method descibed in Annex 2.

    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    P : number or Quantity
        Atmospheric pressure (hPa)
    rho : number or Quantity
        Water vapor density (g/m3)
    T : number or Quantity
        Absolute temperature (K)


    Returns
    -------
    gamma_w : Quantity
        Dry atmosphere specific attenuation (dB/km)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    type_output = type(f)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    P = prepare_quantity(P, u.hPa, 'Atmospheric pressure')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapour density')
    T = prepare_quantity(T, u.K, 'Temperature')
    val = __model.gamma0_approx(f, P, rho, T)
    return prepare_output_array(val, type_output) * u.dB / u.km


def gammaw_exact(f, P, rho, T):
    """
    Method to estimate the specific attenuation due to water vapour using
    the line-by-line method described in Annex 1 of the recommendation.


    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    P : number or Quantity
        Atmospheric pressure (hPa)
    rho : number or Quantity
        Water vapor density (g/m3)
    T : number or Quantity
        Absolute temperature (K)


    Returns
    -------
    gamma_w : Quantity
        Water vapour specific attenuation (dB/km)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    type_output = type(f)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    P = prepare_quantity(P, u.hPa, 'Atmospheric pressure ')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapour density')
    T = prepare_quantity(T, u.K, 'Temperature')
    val = __model.gammaw_exact(f, P, rho, T)
    return prepare_output_array(val, type_output) * u.dB / u.km


def gamma0_exact(f, P, rho, T):
    """
    Method to estimate the specific attenuation due to dry atmosphere using
    the line-by-line method described in Annex 1 of the recommendation.

    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    P : number or Quantity
        Atmospheric pressure (hPa)
    rho : number or Quantity
        Water vapor density (g/m3)
    T : number or Quantity
        Absolute temperature (K)


    Returns
    -------
    gamma_w : Quantity
        Dry atmosphere specific attenuation (dB/km)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    type_output = type(f)
    f = prepare_quantity(f, u.GHz, 'Frequency')
    P = prepare_quantity(P, u.hPa, 'Atmospheric pressure')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapour density')
    T = prepare_quantity(T, u.K, 'Temperature')
    val = __model.gamma0_exact(f, P, rho, T)
    return prepare_output_array(val, type_output) * u.dB / u.km


def gamma_exact(f, P, rho, T):
    """
    Method to estimate the specific attenuation using the line-by-line method
    described in Annex 1 of the recommendation.


    Parameters
    ----------
    f : number or Quantity
        Frequency (GHz)
    P : number or Quantity
        Atmospheric pressure (hPa)
    rho : number or Quantity
        Water vapor density (g/m3)
    T : number or Quantity
        Absolute temperature (K)

    Returns
    -------
    gamma : Quantity
        Specific attenuation (dB/km)

    References
    --------
    [1] Attenuation by atmospheric gases:
    https://www.itu.int/rec/R-REC-P.676/en
    """
    f = prepare_quantity(f, u.GHz, 'Frequency')
    type_output = type(f)

    P = prepare_quantity(P, u.hPa, 'Atmospheric pressure ')
    rho = prepare_quantity(rho, u.g / u.m**3, 'Water vapour density')
    T = prepare_quantity(T, u.K, 'Temperature')
    val = __model.gamma_exact(f, P, rho, T)
    return prepare_output_array(val, type_output) * u.dB / u.km



# -*- coding: utf-8 -*-
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt

import sys
import numpy as np
import unittest as test

import itur
import itur.models.itu676 as itu676
import itur.models.itu835 as itu835


def suite():
    """A test suite for the examples includes in itur. """
    suite = test.TestSuite()

    # Test valid versions
    suite.addTest(TestMapAfrica('test_map_africa'))
    suite.addTest(TestGaseousAttenuation('test_gaseous_attenuation'))
    suite.addTest(TestMultipleLocations('test_multiple_locations'))
    suite.addTest(TestSingleLocation('test_single_location'))
    suite.addTest(TestSingleLocationVsFrequency('test_single_location_vs_f'))
    suite.addTest(TestSingleLocationVsUnavailability(
            'test_single_location_vs_p'))

    return suite


class TestMapAfrica(test.TestCase):

    @staticmethod
    def test_map_africa():
        # Generate a regular grid of latitude and longitudes with 0.1
        # degree resolution for the region of interest.
        lat, lon = itur.utils.regular_lat_lon_grid(lat_max=60,
                                                   lat_min=-60,
                                                   lon_max=65,
                                                   lon_min=-35,
                                                   resolution_lon=1,
                                                   resolution_lat=1)

        # Satellite coordinates (GEO, 4 E)
        lat_sat = 0
        lon_sat = 4
        h_sat = 35786 * itur.u.km

        # Compute the elevation angle between satellite and ground stations
        el = itur.utils.elevation_angle(h_sat, lat_sat, lon_sat, lat, lon)

        # Set the link parameters
        f = 22.5 * itur.u.GHz    # Link frequency
        D = 1.2 * itur.u.m       # Antenna diameters
        p = 0.1                  # Unavailability (Vals exceeded 0.1% of time)

        # Compute the atmospheric attenuation
        Att = itur.atmospheric_attenuation_slant_path(lat, lon, f, el, p, D)

        # Now we show the surface mean temperature distribution
        T = itur.surface_mean_temperature(lat, lon)\
            .to(itur.u.Celsius, equivalencies=itur.u.temperature())

        # Plot the results
        try:
            m = itur.plotting.plot_in_map(Att, lat, lon,
                                       cbar_text='Atmospheric attenuation [dB]',
                                       cmap='magma')

            # Plot the satellite location
            m.scatter(lon_sat, lat_sat, c='white', s=20)

            m = itur.plotting.plot_in_map(
                T, lat, lon, cbar_text='Surface mean temperature [C]',
                cmap='RdBu_r')
        except RuntimeError as e:
            print(e)


class TestMultipleLocations(test.TestCase):

    @staticmethod
    def test_multiple_locations():
        # Obtain the coordinates of the different cities
        cities = {'Boston': (42.36, -71.06),
                  'New York': (40.71, -74.01),
                  'Los Angeles': (34.05, -118.24),
                  'Denver': (39.74, -104.99),
                  'Las Vegas': (36.20, -115.14),
                  'Seattle': (47.61, -122.33),
                  'Washington DC': (38.91, -77.04)}

        lat = [coords[0] for coords in cities.values()]
        lon = [coords[1] for coords in cities.values()]

        # Satellite coordinates (GEO, 4 E)
        lat_sat = 0
        lon_sat = -77
        h_sat = 35786 * itur.u.km

        # Compute the elevation angle between satellite and ground stations
        el = itur.utils.elevation_angle(h_sat, lat_sat, lon_sat, lat, lon)

        # Set the link parameters
        f = 22.5 * itur.u.GHz    # Link frequency
        D = 1.2 * itur.u.m       # Antenna diameters
        p = 0.1                  # Unavailability (Vals exceeded 0.1% of time)

        # Compute the atmospheric attenuation
        Ag, Ac, Ar, As, Att = itur.atmospheric_attenuation_slant_path(
            lat, lon, f, el, p, D, return_contributions=True)

        # Plot the results
        city_idx = np.arange(len(cities))
        width = 0.15

        fig, ax = plt.subplots(1, 1)
        ax.bar(city_idx, Att.value, 0.6, label='Total atmospheric Attenuation')
        ax.bar(city_idx - 1.5 * width, Ar.value, width,
               label='Rain attenuation')
        ax.bar(city_idx - 0.5 * width, Ag.value, width,
               label='Gaseous attenuation')
        ax.bar(city_idx + 0.5 * width, Ac.value, width,
               label='Clouds attenuation')
        ax.bar(city_idx + 1.5 * width, As.value, width,
               label='Scintillation attenuation')

        # Set the labels
        ticks = ax.set_xticklabels([''] + list(cities.keys()))
        for t in ticks:
            t.set_rotation(45)
        ax.set_ylabel('Atmospheric attenuation exceeded for 0.1% [dB]')

        # Format image
        ax.yaxis.grid(which='both', linestyle=':')
        ax.legend(loc='upper center', bbox_to_anchor=(0.5, 1.3), ncol=2)
        plt.tight_layout(rect=(0, 0, 1, 0.85))


class TestSingleLocation(test.TestCase):

    @staticmethod
    def test_single_location():

        # Location of the receiver ground stations
        lat = 41.39
        lon = -71.05

        # Link parameters
        el = 60                # Elevation angle equal to 60 degrees
        f = 22.5 * itur.u.GHz  # Frequency equal to 22.5 GHz
        D = 1 * itur.u.m       # Receiver antenna diameter of 1 m
        p = 0.1                # We compute values exceeded during 0.1 % of
                               # the average year

        # Compute atmospheric parameters
        hs = itur.topographic_altitude(lat, lon)
        T = itur.surface_mean_temperature(lat, lon)
        P = itur.models.itu835.pressure(lat, hs)
        rho_p = itur.surface_water_vapour_density(lat, lon, p, hs)
        itur.models.itu835.water_vapour_density(lat, hs)
        itur.models.itu835.temperature(lat, hs)
        itur.models.itu836.total_water_vapour_content(lat, lon, p, hs)

        # Compute rain and cloud-related parameters
        itur.models.itu618.rain_attenuation_probability(
                lat, lon, el, hs)
        itur.models.itu837.rainfall_probability(lat, lon)
        itur.models.itu837.rainfall_rate(lat, lon, p)
        itur.models.itu839.isoterm_0(lat, lon)
        itur.models.itu839.rain_height(lat, lon)
        itur.models.itu840.columnar_content_reduced_liquid(
                lat, lon, p)
        itur.models.itu676.zenit_water_vapour_attenuation(
                lat, lon, p, f, h=hs)

        # Compute attenuation values
        itur.gaseous_attenuation_slant_path(f, el, rho_p, P, T)
        itur.rain_attenuation(lat, lon, f, el, hs=hs, p=p)
        itur.cloud_attenuation(lat, lon, el, f, p)
        itur.scintillation_attenuation(lat, lon, f, el, p, D)
        itur.atmospheric_attenuation_slant_path(lat, lon, f, el, p, D)


class TestSingleLocationVsFrequency(test.TestCase):

    @staticmethod
    def test_single_location_vs_f():
        # Ground station coordinates (Boston)
        lat_GS = 42.3601
        lon_GS = -71.0942

        ################################################
        # First case: Attenuation vs. frequency        #
        ################################################

        # Satellite coordinates (GEO, 77 W)
        lat_sat = 0
        lon_sat = -77
        h_sat = 35786 * itur.u.km

        # Compute the elevation angle between satellite and ground station
        el = itur.utils.elevation_angle(h_sat, lat_sat, lon_sat,
                                        lat_GS, lon_GS)

        f = 22.5 * itur.u.GHz    # Link frequency
        D = 1.2 * itur.u.m       # Antenna diameters
        p = 1

        f = np.logspace(-0.2, 2, 100) * itur.u.GHz

        Ag, Ac, Ar, As, A =\
            itur.atmospheric_attenuation_slant_path(lat_GS, lon_GS, f,
                                                    el, p, D,
                                                    return_contributions=True)

        # Plot the results
        fig, ax = plt.subplots(1, 1)
        ax.loglog(f, Ag, label='Gaseous attenuation')
        ax.loglog(f, Ac, label='Cloud attenuation')
        ax.loglog(f, Ar, label='Rain attenuation')
        ax.loglog(f, As, label='Scintillation attenuation')
        ax.loglog(f, A, label='Total atmospheric attenuation')

        ax.set_xlabel('Frequency [GHz]')
        ax.set_ylabel('Atmospheric attenuation [dB]')
        ax.grid(which='both', linestyle=':')
        plt.legend()

        ################################################
        # Second case: Attenuation vs. elevation angle #
        ################################################

        f = 22.5 * itur.u.GHz
        el = np.linspace(5, 90, 100)

        Ag, Ac, Ar, As, A =\
            itur.atmospheric_attenuation_slant_path(lat_GS, lon_GS,
                                                    f, el, p, D,
                                                    return_contributions=True)

        # Plot the results
        fig, ax = plt.subplots(1, 1)
        ax.plot(el, Ag, label='Gaseous attenuation')
        ax.plot(el, Ac, label='Cloud attenuation')
        ax.plot(el, Ar, label='Rain attenuation')
        ax.plot(el, As, label='Scintillation attenuation')
        ax.plot(el, A, label='Total atmospheric attenuation')

        ax.set_xlabel('Elevation angle [deg]')
        ax.set_ylabel('Atmospheric attenuation [dB]')
        ax.grid(which='both', linestyle=':')
        plt.legend()


class TestSingleLocationVsUnavailability(test.TestCase):

    @staticmethod
    def test_single_location_vs_p():
        # Ground station coordinates (Boston)
        lat_GS = 42.3601
        lon_GS = -71.0942

        # Satellite coordinates (GEO, 77 W)
        lat_sat = 0
        lon_sat = -77
        h_sat = 35786 * itur.u.km

        # Compute the elevation angle between satellite and ground station
        el = itur.utils.elevation_angle(h_sat, lat_sat, lon_sat,
                                        lat_GS, lon_GS)

        f = 22.5 * itur.u.GHz    # Link frequency
        D = 1.2 * itur.u.m       # Antenna diameters

        # Define unavailabilities vector in logarithmic scale
        p = np.logspace(-1.5, 1.5, 100)

        A_g, A_c, A_r, A_s, A_t = \
            itur.atmospheric_attenuation_slant_path(
                    lat_GS, lon_GS, f, el, p, D, return_contributions=True)

        # Plot the results using matplotlib
        f, ax = plt.subplots(1, 1)
        ax.semilogx(p, A_g.value, label='Gaseous attenuation')
        ax.semilogx(p, A_c.value, label='Cloud attenuation')
        ax.semilogx(p, A_r.value, label='Rain attenuation')
        ax.semilogx(p, A_s.value, label='Scintillation attenuation')
        ax.semilogx(p, A_t.value, label='Total atmospheric attenuation')

        ax.set_xlabel('Percentage of time attenuation value is exceeded [%]')
        ax.set_ylabel('Attenuation [dB]')
        ax.grid(which='both', linestyle=':')
        plt.legend()


class TestGaseousAttenuation(test.TestCase):

    @staticmethod
    def test_gaseous_attenuation():
        # Define atmospheric parameters
        rho_wet = 7.5 * itur.u.g / itur.u.m**3
        rho_dry = 0 * itur.u.g / itur.u.m**3
        P = 1013.25 * itur.u.hPa
        T = 15 * itur.u.deg_C

        # Define frequency logspace parameters
        N_freq = 1000
        fs = np.linspace(0, 1000, N_freq)

        # Compute the attenuation values
        att_wet = itu676.gamma_exact(fs, P, rho_wet, T)
        att_dry = itu676.gamma_exact(fs, P, rho_dry, T)

        # Plot the results
        plt.figure()
        plt.plot(fs, att_wet.value, 'b--', label='Wet atmosphere')
        plt.plot(fs, att_dry.value, 'r', label='Dry atmosphere')
        plt.xlabel('Frequency [GHz]')
        plt.ylabel('Specific attenuation [dB/km]')
        plt.yscale('log')
        plt.xscale('linear')
        plt.xlim(0, 1000)
        plt.ylim(1e-3, 1e5)
        plt.legend()
        plt.grid(which='both', linestyle=':', color='gray',
                 linewidth=0.3, alpha=0.5)
        plt.grid(which='major', linestyle=':', color='black')
        plt.title('FIGURE 1. - Specific attenuation due to atmospheric gases,'
                  '\ncalculated at 1 GHz intervals, including line centres')
        plt.tight_layout()

        #######################################################################
        #               Specific attenuation at different altitudes           #
        #######################################################################

        # Define atmospheric parameters
        hs = np.array([0, 5, 10, 15, 20]) * itur.u.km

        # Define frequency logspace parameters
        N_freq = 2001
        fs = np.linspace(50, 70, N_freq)

        # Plot the results
        plt.figure()

        # Loop over heights and compute values
        for h in hs:
            rho = itu835.standard_water_vapour_density(h)
            P = itu835.standard_pressure(h)
            T = itu835.standard_temperature(h)
            atts = itu676.gamma_exact(fs * itur.u.GHz, P, rho, T)
            plt.plot(fs, atts.value, label='Altitude {0} km'.format(h.value))

        plt.xlabel('Frequency [GHz]')
        plt.ylabel('Specific attenuation [dB/km]')
        plt.yscale('log')
        plt.xscale('linear')
        plt.xlim(50, 70)
        plt.ylim(1e-3, 1e2)
        plt.legend()
        plt.grid(which='both', linestyle=':', color='gray',
                 linewidth=0.3, alpha=0.5)
        plt.grid(which='major', linestyle=':', color='black')
        plt.title('FIGURE 2. - Specific attenuation in the range 50-70 GHz'
                  ' at the\n altitudes indicated, calculated at intervals of'
                  ' 10 MHz\nincluding line centers (0, 5, 10 15, 20) km')
        plt.tight_layout()

        #######################################################################
        #           Comparison of line-by-line and approximate method         #
        #######################################################################
        # Define atmospheric parameters
        el = 90
        rho = 7.5 * itur.u.g / itur.u.m**3
        P = 1013.25 * itur.u.hPa
        T = 15 * itur.u.deg_C

        # Define frequency logspace parameters
        N_freq = 350
        fs = np.linspace(0, 350, N_freq)

        # Initialize result vectors
        atts_approx = []
        atts_exact = []

        # Loop over frequencies and compute values
        atts_approx = itu676.gaseous_attenuation_slant_path(
            fs, el, rho, P, T, mode='approx')

        atts_exact = itu676.gaseous_attenuation_slant_path(
            fs, el, rho, P, T, mode='exact')

        # Plot the results
        plt.figure()
        plt.plot(fs, atts_approx.value, 'b--',
                 label='Approximate method Annex 2')
        plt.plot(fs, atts_exact.value, 'r', label='Exact line-by-line method')
        plt.xlabel('Frequency [GHz]')
        plt.ylabel('Attenuation [dB]')
        plt.yscale('log')
        plt.xscale('log')
        plt.legend()
        plt.grid(which='both', linestyle=':', color='gray',
                 linewidth=0.3, alpha=0.5)
        plt.grid(which='major', linestyle=':', color='black')
        plt.title('Comparison of line-by-line method to approximate method')
        plt.tight_layout()


if __name__ == '__main__':
    suite = suite()
    print('Test examples of the code')
    print('------------------------')
    print(
        'A total of %d test-cases are going to be tested' %
        suite.countTestCases())
    sys.stdout.flush()
    test.TextTestRunner(verbosity=2).run(suite)
